import logging

from defi_services.constants.chain_constant import Chain
from defi_services.constants.entities.lending_constant import Lending
from defi_services.jobs.queriers.state_querier import StateQuerier
from defi_services.services.lending.lending_info.fantom.geist_ftm import GEIST_ETH
from defi_services.services.lending.valas_services import ValasStateService

logger = logging.getLogger("Geist Lending Pool State Service")


class GeistInfo:
    mapping = {
        Chain.fantom: GEIST_ETH
    }


class GeistStateService(ValasStateService):
    def __init__(self, state_service: StateQuerier, chain_id: str = "0xfa"):
        super().__init__(state_service, chain_id)
        self.name = f"{chain_id}_{Lending.geist}"
        self.pool_info = GeistInfo.mapping.get(chain_id)

    def get_service_info(self):
        info = {
            Lending.geist: {
                "chain_id": self.chain_id,
                "type": "lending",
                "protocol_info": self.pool_info
            }
        }
        return info



import json

LIDO_ALLOWED_RECIPIENTS_FACTORY_ABI = json.loads('''
[
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "creator",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "addAllowedRecipient",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "address",
        "name": "trustedCaller",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "address",
        "name": "allowedRecipientsRegistry",
        "type": "address"
      }
    ],
    "name": "AddAllowedRecipientDeployed",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "creator",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "allowedRecipientsRegistry",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "address",
        "name": "_defaultAdmin",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "address[]",
        "name": "addRecipientToAllowedListRoleHolders",
        "type": "address[]"
      },
      {
        "indexed": false,
        "internalType": "address[]",
        "name": "removeRecipientFromAllowedListRoleHolders",
        "type": "address[]"
      },
      {
        "indexed": false,
        "internalType": "address[]",
        "name": "setLimitParametersRoleHolders",
        "type": "address[]"
      },
      {
        "indexed": false,
        "internalType": "address[]",
        "name": "updateSpentAmountRoleHolders",
        "type": "address[]"
      },
      {
        "indexed": false,
        "internalType": "contract IBokkyPooBahsDateTimeContract",
        "name": "bokkyPooBahsDateTimeContract",
        "type": "address"
      }
    ],
    "name": "AllowedRecipientsRegistryDeployed",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "creator",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "removeAllowedRecipient",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "address",
        "name": "trustedCaller",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "address",
        "name": "allowedRecipientsRegistry",
        "type": "address"
      }
    ],
    "name": "RemoveAllowedRecipientDeployed",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "creator",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "topUpAllowedRecipients",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "address",
        "name": "trustedCaller",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "address",
        "name": "allowedRecipientsRegistry",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "address",
        "name": "finance",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "address",
        "name": "token",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "address",
        "name": "easyTrack",
        "type": "address"
      }
    ],
    "name": "TopUpAllowedRecipientsDeployed",
    "type": "event"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_trustedCaller",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_allowedRecipientsRegistry",
        "type": "address"
      }
    ],
    "name": "deployAddAllowedRecipient",
    "outputs": [
      {
        "internalType": "contract AddAllowedRecipient",
        "name": "addAllowedRecipient",
        "type": "address"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_defaultAdmin",
        "type": "address"
      },
      {
        "internalType": "address[]",
        "name": "_addRecipientToAllowedListRoleHolders",
        "type": "address[]"
      },
      {
        "internalType": "address[]",
        "name": "_removeRecipientFromAllowedListRoleHolders",
        "type": "address[]"
      },
      {
        "internalType": "address[]",
        "name": "_setLimitParametersRoleHolders",
        "type": "address[]"
      },
      {
        "internalType": "address[]",
        "name": "_updateSpentAmountRoleHolders",
        "type": "address[]"
      },
      {
        "internalType": "contract IBokkyPooBahsDateTimeContract",
        "name": "_bokkyPooBahsDateTimeContract",
        "type": "address"
      }
    ],
    "name": "deployAllowedRecipientsRegistry",
    "outputs": [
      {
        "internalType": "contract AllowedRecipientsRegistry",
        "name": "registry",
        "type": "address"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_trustedCaller",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_allowedRecipientsRegistry",
        "type": "address"
      }
    ],
    "name": "deployRemoveAllowedRecipient",
    "outputs": [
      {
        "internalType": "contract RemoveAllowedRecipient",
        "name": "removeAllowedRecipient",
        "type": "address"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_trustedCaller",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_allowedRecipientsRegistry",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_token",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_finance",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_easyTrack",
        "type": "address"
      }
    ],
    "name": "deployTopUpAllowedRecipients",
    "outputs": [
      {
        "internalType": "contract TopUpAllowedRecipients",
        "name": "topUpAllowedRecipients",
        "type": "address"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  }
]
''')



import json

DEFAULT_RESERVE_INTEREST_RATE_STRATEGY = json.loads("""
[
    {
      "inputs": [
        {
          "internalType": "contract ILendingPoolAddressesProvider",
          "name": "provider",
          "type": "address"
        },
        {
          "internalType": "uint256",
          "name": "optimalUtilizationRate",
          "type": "uint256"
        },
        {
          "internalType": "uint256",
          "name": "baseVariableBorrowRate",
          "type": "uint256"
        },
        {
          "internalType": "uint256",
          "name": "variableRateSlope1",
          "type": "uint256"
        },
        {
          "internalType": "uint256",
          "name": "variableRateSlope2",
          "type": "uint256"
        },
        {
          "internalType": "uint256",
          "name": "stableRateSlope1",
          "type": "uint256"
        },
        {
          "internalType": "uint256",
          "name": "stableRateSlope2",
          "type": "uint256"
        }
      ],
      "stateMutability": "nonpayable",
      "type": "constructor"
    },
    {
      "inputs": [],
      "name": "EXCESS_UTILIZATION_RATE",
      "outputs": [
        {
          "internalType": "uint256",
          "name": "",
          "type": "uint256"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [],
      "name": "OPTIMAL_UTILIZATION_RATE",
      "outputs": [
        {
          "internalType": "uint256",
          "name": "",
          "type": "uint256"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [],
      "name": "addressesProvider",
      "outputs": [
        {
          "internalType": "contract ILendingPoolAddressesProvider",
          "name": "",
          "type": "address"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [],
      "name": "baseVariableBorrowRate",
      "outputs": [
        {
          "internalType": "uint256",
          "name": "",
          "type": "uint256"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [
        {
          "internalType": "address",
          "name": "reserve",
          "type": "address"
        },
        {
          "internalType": "uint256",
          "name": "availableLiquidity",
          "type": "uint256"
        },
        {
          "internalType": "uint256",
          "name": "totalStableDebt",
          "type": "uint256"
        },
        {
          "internalType": "uint256",
          "name": "totalVariableDebt",
          "type": "uint256"
        },
        {
          "internalType": "uint256",
          "name": "averageStableBorrowRate",
          "type": "uint256"
        },
        {
          "internalType": "uint256",
          "name": "reserveFactor",
          "type": "uint256"
        }
      ],
      "name": "calculateInterestRates",
      "outputs": [
        {
          "internalType": "uint256",
          "name": "",
          "type": "uint256"
        },
        {
          "internalType": "uint256",
          "name": "",
          "type": "uint256"
        },
        {
          "internalType": "uint256",
          "name": "",
          "type": "uint256"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [],
      "name": "getMaxVariableBorrowRate",
      "outputs": [
        {
          "internalType": "uint256",
          "name": "",
          "type": "uint256"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [],
      "name": "stableRateSlope1",
      "outputs": [
        {
          "internalType": "uint256",
          "name": "",
          "type": "uint256"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [],
      "name": "stableRateSlope2",
      "outputs": [
        {
          "internalType": "uint256",
          "name": "",
          "type": "uint256"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [],
      "name": "variableRateSlope1",
      "outputs": [
        {
          "internalType": "uint256",
          "name": "",
          "type": "uint256"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [],
      "name": "variableRateSlope2",
      "outputs": [
        {
          "internalType": "uint256",
          "name": "",
          "type": "uint256"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    }
]
""")
        



import json

LENDING_POOL = json.loads("""
[
    {
      "anonymous": false,
      "inputs": [
        {
          "indexed": true,
          "internalType": "address",
          "name": "reserve",
          "type": "address"
        },
        {
          "indexed": false,
          "internalType": "address",
          "name": "user",
          "type": "address"
        },
        {
          "indexed": true,
          "internalType": "address",
          "name": "onBehalfOf",
          "type": "address"
        },
        {
          "indexed": false,
          "internalType": "uint256",
          "name": "amount",
          "type": "uint256"
        },
        {
          "indexed": false,
          "internalType": "uint256",
          "name": "borrowRateMode",
          "type": "uint256"
        },
        {
          "indexed": false,
          "internalType": "uint256",
          "name": "borrowRate",
          "type": "uint256"
        },
        {
          "indexed": true,
          "internalType": "uint16",
          "name": "referral",
          "type": "uint16"
        }
      ],
      "name": "Borrow",
      "type": "event"
    },
    {
      "anonymous": false,
      "inputs": [
        {
          "indexed": true,
          "internalType": "address",
          "name": "reserve",
          "type": "address"
        },
        {
          "indexed": false,
          "internalType": "address",
          "name": "user",
          "type": "address"
        },
        {
          "indexed": true,
          "internalType": "address",
          "name": "onBehalfOf",
          "type": "address"
        },
        {
          "indexed": false,
          "internalType": "uint256",
          "name": "amount",
          "type": "uint256"
        },
        {
          "indexed": true,
          "internalType": "uint16",
          "name": "referral",
          "type": "uint16"
        }
      ],
      "name": "Deposit",
      "type": "event"
    },
    {
      "anonymous": false,
      "inputs": [
        {
          "indexed": true,
          "internalType": "address",
          "name": "target",
          "type": "address"
        },
        {
          "indexed": true,
          "internalType": "address",
          "name": "initiator",
          "type": "address"
        },
        {
          "indexed": true,
          "internalType": "address",
          "name": "asset",
          "type": "address"
        },
        {
          "indexed": false,
          "internalType": "uint256",
          "name": "amount",
          "type": "uint256"
        },
        {
          "indexed": false,
          "internalType": "uint256",
          "name": "premium",
          "type": "uint256"
        },
        {
          "indexed": false,
          "internalType": "uint16",
          "name": "referralCode",
          "type": "uint16"
        }
      ],
      "name": "FlashLoan",
      "type": "event"
    },
    {
      "anonymous": false,
      "inputs": [
        {
          "indexed": true,
          "internalType": "address",
          "name": "collateralAsset",
          "type": "address"
        },
        {
          "indexed": true,
          "internalType": "address",
          "name": "debtAsset",
          "type": "address"
        },
        {
          "indexed": true,
          "internalType": "address",
          "name": "user",
          "type": "address"
        },
        {
          "indexed": false,
          "internalType": "uint256",
          "name": "debtToCover",
          "type": "uint256"
        },
        {
          "indexed": false,
          "internalType": "uint256",
          "name": "liquidatedCollateralAmount",
          "type": "uint256"
        },
        {
          "indexed": false,
          "internalType": "address",
          "name": "liquidator",
          "type": "address"
        },
        {
          "indexed": false,
          "internalType": "bool",
          "name": "receiveAToken",
          "type": "bool"
        }
      ],
      "name": "LiquidationCall",
      "type": "event"
    },
    {
      "anonymous": false,
      "inputs": [],
      "name": "Paused",
      "type": "event"
    },
    {
      "anonymous": false,
      "inputs": [
        {
          "indexed": true,
          "internalType": "address",
          "name": "reserve",
          "type": "address"
        },
        {
          "indexed": true,
          "internalType": "address",
          "name": "user",
          "type": "address"
        }
      ],
      "name": "RebalanceStableBorrowRate",
      "type": "event"
    },
    {
      "anonymous": false,
      "inputs": [
        {
          "indexed": true,
          "internalType": "address",
          "name": "reserve",
          "type": "address"
        },
        {
          "indexed": true,
          "internalType": "address",
          "name": "user",
          "type": "address"
        },
        {
          "indexed": true,
          "internalType": "address",
          "name": "repayer",
          "type": "address"
        },
        {
          "indexed": false,
          "internalType": "uint256",
          "name": "amount",
          "type": "uint256"
        }
      ],
      "name": "Repay",
      "type": "event"
    },
    {
      "anonymous": false,
      "inputs": [
        {
          "indexed": true,
          "internalType": "address",
          "name": "reserve",
          "type": "address"
        },
        {
          "indexed": false,
          "internalType": "uint256",
          "name": "liquidityRate",
          "type": "uint256"
        },
        {
          "indexed": false,
          "internalType": "uint256",
          "name": "stableBorrowRate",
          "type": "uint256"
        },
        {
          "indexed": false,
          "internalType": "uint256",
          "name": "variableBorrowRate",
          "type": "uint256"
        },
        {
          "indexed": false,
          "internalType": "uint256",
          "name": "liquidityIndex",
          "type": "uint256"
        },
        {
          "indexed": false,
          "internalType": "uint256",
          "name": "variableBorrowIndex",
          "type": "uint256"
        }
      ],
      "name": "ReserveDataUpdated",
      "type": "event"
    },
    {
      "anonymous": false,
      "inputs": [
        {
          "indexed": true,
          "internalType": "address",
          "name": "reserve",
          "type": "address"
        },
        {
          "indexed": true,
          "internalType": "address",
          "name": "user",
          "type": "address"
        }
      ],
      "name": "ReserveUsedAsCollateralDisabled",
      "type": "event"
    },
    {
      "anonymous": false,
      "inputs": [
        {
          "indexed": true,
          "internalType": "address",
          "name": "reserve",
          "type": "address"
        },
        {
          "indexed": true,
          "internalType": "address",
          "name": "user",
          "type": "address"
        }
      ],
      "name": "ReserveUsedAsCollateralEnabled",
      "type": "event"
    },
    {
      "anonymous": false,
      "inputs": [
        {
          "indexed": true,
          "internalType": "address",
          "name": "reserve",
          "type": "address"
        },
        {
          "indexed": true,
          "internalType": "address",
          "name": "user",
          "type": "address"
        },
        {
          "indexed": false,
          "internalType": "uint256",
          "name": "rateMode",
          "type": "uint256"
        }
      ],
      "name": "Swap",
      "type": "event"
    },
    {
      "anonymous": false,
      "inputs": [],
      "name": "Unpaused",
      "type": "event"
    },
    {
      "anonymous": false,
      "inputs": [
        {
          "indexed": true,
          "internalType": "address",
          "name": "reserve",
          "type": "address"
        },
        {
          "indexed": true,
          "internalType": "address",
          "name": "user",
          "type": "address"
        },
        {
          "indexed": true,
          "internalType": "address",
          "name": "to",
          "type": "address"
        },
        {
          "indexed": false,
          "internalType": "uint256",
          "name": "amount",
          "type": "uint256"
        }
      ],
      "name": "Withdraw",
      "type": "event"
    },
    {
      "inputs": [],
      "name": "FLASHLOAN_PREMIUM_TOTAL",
      "outputs": [
        {
          "internalType": "uint256",
          "name": "",
          "type": "uint256"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [],
      "name": "LENDINGPOOL_REVISION",
      "outputs": [
        {
          "internalType": "uint256",
          "name": "",
          "type": "uint256"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [],
      "name": "MAX_NUMBER_RESERVES",
      "outputs": [
        {
          "internalType": "uint256",
          "name": "",
          "type": "uint256"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [],
      "name": "MAX_STABLE_RATE_BORROW_SIZE_PERCENT",
      "outputs": [
        {
          "internalType": "uint256",
          "name": "",
          "type": "uint256"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [
        {
          "internalType": "address",
          "name": "asset",
          "type": "address"
        },
        {
          "internalType": "uint256",
          "name": "amount",
          "type": "uint256"
        },
        {
          "internalType": "uint256",
          "name": "interestRateMode",
          "type": "uint256"
        },
        {
          "internalType": "uint16",
          "name": "referralCode",
          "type": "uint16"
        },
        {
          "internalType": "address",
          "name": "onBehalfOf",
          "type": "address"
        }
      ],
      "name": "borrow",
      "outputs": [],
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "inputs": [
        {
          "internalType": "address",
          "name": "asset",
          "type": "address"
        },
        {
          "internalType": "uint256",
          "name": "amount",
          "type": "uint256"
        },
        {
          "internalType": "address",
          "name": "onBehalfOf",
          "type": "address"
        },
        {
          "internalType": "uint16",
          "name": "referralCode",
          "type": "uint16"
        }
      ],
      "name": "deposit",
      "outputs": [],
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "inputs": [
        {
          "internalType": "address",
          "name": "asset",
          "type": "address"
        },
        {
          "internalType": "address",
          "name": "from",
          "type": "address"
        },
        {
          "internalType": "address",
          "name": "to",
          "type": "address"
        },
        {
          "internalType": "uint256",
          "name": "amount",
          "type": "uint256"
        },
        {
          "internalType": "uint256",
          "name": "balanceFromBefore",
          "type": "uint256"
        },
        {
          "internalType": "uint256",
          "name": "balanceToBefore",
          "type": "uint256"
        }
      ],
      "name": "finalizeTransfer",
      "outputs": [],
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "inputs": [
        {
          "internalType": "address",
          "name": "receiverAddress",
          "type": "address"
        },
        {
          "internalType": "address[]",
          "name": "assets",
          "type": "address[]"
        },
        {
          "internalType": "uint256[]",
          "name": "amounts",
          "type": "uint256[]"
        },
        {
          "internalType": "uint256[]",
          "name": "modes",
          "type": "uint256[]"
        },
        {
          "internalType": "address",
          "name": "onBehalfOf",
          "type": "address"
        },
        {
          "internalType": "bytes",
          "name": "params",
          "type": "bytes"
        },
        {
          "internalType": "uint16",
          "name": "referralCode",
          "type": "uint16"
        }
      ],
      "name": "flashLoan",
      "outputs": [],
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "inputs": [],
      "name": "getAddressesProvider",
      "outputs": [
        {
          "internalType": "contract ILendingPoolAddressesProvider",
          "name": "",
          "type": "address"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [
        {
          "internalType": "address",
          "name": "asset",
          "type": "address"
        }
      ],
      "name": "getConfiguration",
      "outputs": [
        {
          "components": [
            {
              "internalType": "uint256",
              "name": "data",
              "type": "uint256"
            }
          ],
          "internalType": "struct DataTypes.ReserveConfigurationMap",
          "name": "",
          "type": "tuple"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [
        {
          "internalType": "address",
          "name": "asset",
          "type": "address"
        }
      ],
      "name": "getReserveData",
      "outputs": [
        {
          "components": [
            {
              "components": [
                {
                  "internalType": "uint256",
                  "name": "data",
                  "type": "uint256"
                }
              ],
              "internalType": "struct DataTypes.ReserveConfigurationMap",
              "name": "configuration",
              "type": "tuple"
            },
            {
              "internalType": "uint128",
              "name": "liquidityIndex",
              "type": "uint128"
            },
            {
              "internalType": "uint128",
              "name": "variableBorrowIndex",
              "type": "uint128"
            },
            {
              "internalType": "uint128",
              "name": "currentLiquidityRate",
              "type": "uint128"
            },
            {
              "internalType": "uint128",
              "name": "currentVariableBorrowRate",
              "type": "uint128"
            },
            {
              "internalType": "uint128",
              "name": "currentStableBorrowRate",
              "type": "uint128"
            },
            {
              "internalType": "uint40",
              "name": "lastUpdateTimestamp",
              "type": "uint40"
            },
            {
              "internalType": "address",
              "name": "aTokenAddress",
              "type": "address"
            },
            {
              "internalType": "address",
              "name": "stableDebtTokenAddress",
              "type": "address"
            },
            {
              "internalType": "address",
              "name": "variableDebtTokenAddress",
              "type": "address"
            },
            {
              "internalType": "address",
              "name": "interestRateStrategyAddress",
              "type": "address"
            },
            {
              "internalType": "uint8",
              "name": "id",
              "type": "uint8"
            }
          ],
          "internalType": "struct DataTypes.ReserveData",
          "name": "",
          "type": "tuple"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [
        {
          "internalType": "address",
          "name": "asset",
          "type": "address"
        }
      ],
      "name": "getReserveNormalizedIncome",
      "outputs": [
        {
          "internalType": "uint256",
          "name": "",
          "type": "uint256"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [
        {
          "internalType": "address",
          "name": "asset",
          "type": "address"
        }
      ],
      "name": "getReserveNormalizedVariableDebt",
      "outputs": [
        {
          "internalType": "uint256",
          "name": "",
          "type": "uint256"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [],
      "name": "getReservesList",
      "outputs": [
        {
          "internalType": "address[]",
          "name": "",
          "type": "address[]"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [
        {
          "internalType": "address",
          "name": "user",
          "type": "address"
        }
      ],
      "name": "getUserAccountData",
      "outputs": [
        {
          "internalType": "uint256",
          "name": "totalCollateralETH",
          "type": "uint256"
        },
        {
          "internalType": "uint256",
          "name": "totalDebtETH",
          "type": "uint256"
        },
        {
          "internalType": "uint256",
          "name": "availableBorrowsETH",
          "type": "uint256"
        },
        {
          "internalType": "uint256",
          "name": "currentLiquidationThreshold",
          "type": "uint256"
        },
        {
          "internalType": "uint256",
          "name": "ltv",
          "type": "uint256"
        },
        {
          "internalType": "uint256",
          "name": "healthFactor",
          "type": "uint256"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [
        {
          "internalType": "address",
          "name": "user",
          "type": "address"
        }
      ],
      "name": "getUserConfiguration",
      "outputs": [
        {
          "components": [
            {
              "internalType": "uint256",
              "name": "data",
              "type": "uint256"
            }
          ],
          "internalType": "struct DataTypes.UserConfigurationMap",
          "name": "",
          "type": "tuple"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [
        {
          "internalType": "address",
          "name": "asset",
          "type": "address"
        },
        {
          "internalType": "address",
          "name": "aTokenAddress",
          "type": "address"
        },
        {
          "internalType": "address",
          "name": "stableDebtAddress",
          "type": "address"
        },
        {
          "internalType": "address",
          "name": "variableDebtAddress",
          "type": "address"
        },
        {
          "internalType": "address",
          "name": "interestRateStrategyAddress",
          "type": "address"
        }
      ],
      "name": "initReserve",
      "outputs": [],
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "inputs": [
        {
          "internalType": "contract ILendingPoolAddressesProvider",
          "name": "provider",
          "type": "address"
        }
      ],
      "name": "initialize",
      "outputs": [],
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "inputs": [
        {
          "internalType": "address",
          "name": "collateralAsset",
          "type": "address"
        },
        {
          "internalType": "address",
          "name": "debtAsset",
          "type": "address"
        },
        {
          "internalType": "address",
          "name": "user",
          "type": "address"
        },
        {
          "internalType": "uint256",
          "name": "debtToCover",
          "type": "uint256"
        },
        {
          "internalType": "bool",
          "name": "receiveAToken",
          "type": "bool"
        }
      ],
      "name": "liquidationCall",
      "outputs": [],
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "inputs": [],
      "name": "paused",
      "outputs": [
        {
          "internalType": "bool",
          "name": "",
          "type": "bool"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [
        {
          "internalType": "address",
          "name": "asset",
          "type": "address"
        },
        {
          "internalType": "address",
          "name": "user",
          "type": "address"
        }
      ],
      "name": "rebalanceStableBorrowRate",
      "outputs": [],
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "inputs": [
        {
          "internalType": "address",
          "name": "asset",
          "type": "address"
        },
        {
          "internalType": "uint256",
          "name": "amount",
          "type": "uint256"
        },
        {
          "internalType": "uint256",
          "name": "rateMode",
          "type": "uint256"
        },
        {
          "internalType": "address",
          "name": "onBehalfOf",
          "type": "address"
        }
      ],
      "name": "repay",
      "outputs": [
        {
          "internalType": "uint256",
          "name": "",
          "type": "uint256"
        }
      ],
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "inputs": [
        {
          "internalType": "address",
          "name": "asset",
          "type": "address"
        },
        {
          "internalType": "uint256",
          "name": "configuration",
          "type": "uint256"
        }
      ],
      "name": "setConfiguration",
      "outputs": [],
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "inputs": [
        {
          "internalType": "bool",
          "name": "val",
          "type": "bool"
        }
      ],
      "name": "setPause",
      "outputs": [],
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "inputs": [
        {
          "internalType": "address",
          "name": "asset",
          "type": "address"
        },
        {
          "internalType": "address",
          "name": "rateStrategyAddress",
          "type": "address"
        }
      ],
      "name": "setReserveInterestRateStrategyAddress",
      "outputs": [],
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "inputs": [
        {
          "internalType": "address",
          "name": "asset",
          "type": "address"
        },
        {
          "internalType": "bool",
          "name": "useAsCollateral",
          "type": "bool"
        }
      ],
      "name": "setUserUseReserveAsCollateral",
      "outputs": [],
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "inputs": [
        {
          "internalType": "address",
          "name": "asset",
          "type": "address"
        },
        {
          "internalType": "uint256",
          "name": "rateMode",
          "type": "uint256"
        }
      ],
      "name": "swapBorrowRateMode",
      "outputs": [],
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "inputs": [
        {
          "internalType": "address",
          "name": "asset",
          "type": "address"
        },
        {
          "internalType": "uint256",
          "name": "amount",
          "type": "uint256"
        },
        {
          "internalType": "address",
          "name": "to",
          "type": "address"
        }
      ],
      "name": "withdraw",
      "outputs": [
        {
          "internalType": "uint256",
          "name": "",
          "type": "uint256"
        }
      ],
      "stateMutability": "nonpayable",
      "type": "function"
    }
]
""")
        


import json

from defi_services.abis.lending.venus.abi.comptroller_defi import COMPTROLLER_DEFI
from defi_services.abis.lending.venus.abi.comptrollerbeacon import COMPTROLLERBEACON
from defi_services.abis.lending.venus.abi.comptrollerimpl import COMPTROLLERIMPL
from defi_services.abis.lending.venus.abi.defaultproxyadmin import DEFAULTPROXYADMIN
from defi_services.abis.lending.venus.abi.jumpratemodelv2 import JUMPRATEMODELV2
from defi_services.abis.lending.venus.abi.poollens import POOLLENS
from defi_services.abis.lending.venus.abi.poolregistry import POOLREGISTRY
from defi_services.abis.lending.venus.abi.poolregistry_implementation import POOLREGISTRY_IMPLEMENTATION
from defi_services.abis.lending.venus.abi.poolregistry_proxy import POOLREGISTRY_PROXY
from defi_services.abis.lending.venus.abi.protocolsharereserve import PROTOCOLSHARERESERVE
from defi_services.abis.lending.venus.abi.protocolsharereserve_implementation import PROTOCOLSHARERESERVE_IMPLEMENTATION
from defi_services.abis.lending.venus.abi.protocolsharereserve_proxy import PROTOCOLSHARERESERVE_PROXY
from defi_services.abis.lending.venus.abi.rewardsdistributor_defi import REWARDSDISTRIBUTOR_DEFI_0
from defi_services.abis.lending.venus.abi.rewardsdistributorimpl import REWARDSDISTRIBUTORIMPL
from defi_services.abis.lending.venus.abi.riskfund import RISKFUND
from defi_services.abis.lending.venus.abi.riskfund_implementation import RISKFUND_IMPLEMENTATION
from defi_services.abis.lending.venus.abi.riskfund_proxy import RISKFUND_PROXY
from defi_services.abis.lending.venus.abi.shortfall import SHORTFALL
from defi_services.abis.lending.venus.abi.shortfall_implementation import SHORTFALL_IMPLEMENTATION
from defi_services.abis.lending.venus.abi.shortfall_proxy import SHORTFALL_PROXY
from defi_services.abis.lending.venus.abi.swaprouter_defi import SWAPROUTER_DEFI
from defi_services.abis.lending.venus.abi.vtokenbeacon import VTOKENBEACON
from defi_services.abis.lending.venus.abi.vtokenimpl import VTOKENIMPL
from defi_services.abis.lending.venus.governance.access_control_manager import ACCESS_CONTROL_MANAGER
from defi_services.abis.lending.venus.governance.access_controlled_v5 import ACCESS_CONTROLLED_V5
from defi_services.abis.lending.venus.governance.access_controlled_v8 import ACCESS_CONTROLLED_V8
from defi_services.abis.lending.venus.governance.governor_bravo_delegate import GOVERNOR_BRAVO_DELEGATE
from defi_services.abis.lending.venus.governance.governor_bravo_delegator import GOVERNOR_BRAVO_DELEGATOR
from defi_services.abis.lending.venus.governance.governor_bravo_events import GOVERNOR_BRAVO_EVENTS
from defi_services.abis.lending.venus.governance.timelock import TIMELOCK
from defi_services.abis.lending.venus.venus.admin.v_b_n_b_admin import V_B_N_B_ADMIN
from defi_services.abis.lending.venus.venus.comptroller.diamond import DIAMOND
from defi_services.abis.lending.venus.venus.comptroller.diamond_consolidated import DIAMOND_CONSOLIDATED
from defi_services.abis.lending.venus.venus.comptroller.facet_base import FACET_BASE
from defi_services.abis.lending.venus.venus.comptroller.market_facet import MARKET_FACET
from defi_services.abis.lending.venus.venus.comptroller.policy_facet import POLICY_FACET
from defi_services.abis.lending.venus.venus.comptroller.reward_facet import REWARD_FACET
from defi_services.abis.lending.venus.venus.comptroller.setter_facet import SETTER_FACET
from defi_services.abis.lending.venus.venus.comptroller.unitroller import UNITROLLER
from defi_services.abis.lending.venus.venus.comptroller.x_v_s_rewards_helper import X_V_S_REWARDS_HELPER
from defi_services.abis.lending.venus.venus.delegateborrowers.swap_debt_delegate import SWAP_DEBT_DELEGATE
from defi_services.abis.lending.venus.venus.governance.v_treasury import V_TREASURY
from defi_services.abis.lending.venus.venus.interestratemodels.interest_rate_model import INTEREST_RATE_MODEL
from defi_services.abis.lending.venus.venus.interestratemodels.jump_rate_model import JUMP_RATE_MODEL
from defi_services.abis.lending.venus.venus.interestratemodels.white_paper_interest_rate_model import \
    WHITE_PAPER_INTEREST_RATE_MODEL
from defi_services.abis.lending.venus.venus.lens.comptroller_lens import COMPTROLLER_LENS
from defi_services.abis.lending.venus.venus.lens.snapshot_lens import SNAPSHOT_LENS
from defi_services.abis.lending.venus.venus.lens.venus_lens import VENUS_LENS
from defi_services.abis.lending.venus.venus.lens.x_v_s_staking_lens import X_V_S_STAKING_LENS
from defi_services.abis.lending.venus.venus.liquidator.b_u_s_d_liquidator import B_U_S_D_LIQUIDATOR
from defi_services.abis.lending.venus.venus.liquidator.liquidator import LIQUIDATOR
from defi_services.abis.lending.venus.venus.oracle.price_oracle import PRICE_ORACLE
from defi_services.abis.lending.venus.venus.pegstability.peg_stability import PEG_STABILITY
from defi_services.abis.lending.venus.venus.tokens.v_a_i import V_A_I
from defi_services.abis.lending.venus.venus.tokens.v_a_i_controller import V_A_I_CONTROLLER
from defi_services.abis.lending.venus.venus.tokens.v_a_i_unitroller import V_A_I_UNITROLLER
from defi_services.abis.lending.venus.venus.tokens.v_bep20 import V_BEP20
from defi_services.abis.lending.venus.venus.tokens.v_bep20_delegate import V_BEP20_DELEGATE
from defi_services.abis.lending.venus.venus.tokens.v_bep20_delegator import V_BEP20_DELEGATOR
from defi_services.abis.lending.venus.venus.tokens.v_bep20_immutable import V_BEP20_IMMUTABLE
from defi_services.abis.lending.venus.venus.tokens.v_r_t import V_R_T
from defi_services.abis.lending.venus.venus.tokens.v_r_t_converter import V_R_T_CONVERTER
from defi_services.abis.lending.venus.venus.tokens.v_r_t_converter_proxy import V_R_T_CONVERTER_PROXY
from defi_services.abis.lending.venus.venus.tokens.v_token import V_TOKEN
from defi_services.abis.lending.venus.venus.tokens.x_v_s import X_V_S
from defi_services.abis.lending.venus.venus.tokens.x_v_s_vesting import X_V_S_VESTING
from defi_services.abis.lending.venus.venus.tokens.x_v_s_vesting_proxy import X_V_S_VESTING_PROXY
from defi_services.abis.lending.venus.venus.vault.v_a_i_vault import V_A_I_VAULT
from defi_services.abis.lending.venus.venus.vault.v_a_i_vault_proxy import V_A_I_VAULT_PROXY
from defi_services.abis.lending.venus.venus.vrtvault.v_r_t_vault import V_R_T_VAULT
from defi_services.abis.lending.venus.venus.vrtvault.v_r_t_vault_proxy import V_R_T_VAULT_PROXY
from defi_services.abis.lending.venus.venus.xvsvault.x_v_s_vault import X_V_S_VAULT
from defi_services.abis.lending.venus.venus.xvsvault.x_v_s_vault_error_reporter import X_V_S_VAULT_ERROR_REPORTER
from defi_services.abis.lending.venus.venus.xvsvault.x_v_s_vault_proxy import X_V_S_VAULT_PROXY
from defi_services.abis.lending.venus.venus_comptroller_abi import VENUS_COMPTROLLER_ABI


class VENUS_ABIS:
    mapping = [
        VENUS_COMPTROLLER_ABI,
        COMPTROLLER_DEFI,
        COMPTROLLERBEACON,
        COMPTROLLERIMPL,
        DEFAULTPROXYADMIN,
        JUMPRATEMODELV2,
        POOLLENS,
        POOLREGISTRY,
        POOLREGISTRY_IMPLEMENTATION,
        POOLREGISTRY_PROXY,
        PROTOCOLSHARERESERVE,
        PROTOCOLSHARERESERVE_IMPLEMENTATION,
        PROTOCOLSHARERESERVE_PROXY,
        REWARDSDISTRIBUTOR_DEFI_0,
        REWARDSDISTRIBUTORIMPL,
        RISKFUND,
        RISKFUND_IMPLEMENTATION,
        RISKFUND_PROXY,
        SHORTFALL,
        SHORTFALL_IMPLEMENTATION,
        SHORTFALL_PROXY,
        SWAPROUTER_DEFI,
        VTOKENBEACON,
        VTOKENIMPL,
        ACCESS_CONTROL_MANAGER,
        ACCESS_CONTROLLED_V5,
        ACCESS_CONTROLLED_V8,
        GOVERNOR_BRAVO_DELEGATOR,
        GOVERNOR_BRAVO_DELEGATE,
        GOVERNOR_BRAVO_EVENTS,
        TIMELOCK,
        V_B_N_B_ADMIN,
        DIAMOND,
        DIAMOND_CONSOLIDATED,
        FACET_BASE,
        MARKET_FACET,
        POLICY_FACET,
        REWARD_FACET,
        SETTER_FACET,
        UNITROLLER,
        X_V_S_REWARDS_HELPER,
        SWAP_DEBT_DELEGATE,
        V_TREASURY,
        INTEREST_RATE_MODEL,
        JUMP_RATE_MODEL,
        WHITE_PAPER_INTEREST_RATE_MODEL,
        COMPTROLLER_LENS,
        SNAPSHOT_LENS,
        VENUS_LENS,
        X_V_S_STAKING_LENS,
        B_U_S_D_LIQUIDATOR,
        LIQUIDATOR,
        PRICE_ORACLE,
        PEG_STABILITY,
        V_A_I,
        V_A_I_CONTROLLER,
        V_A_I_UNITROLLER,
        V_BEP20,
        V_BEP20_DELEGATE,
        V_BEP20_DELEGATOR,
        V_BEP20_IMMUTABLE,
        V_R_T,
        V_R_T_CONVERTER,
        V_R_T_CONVERTER_PROXY,
        V_TOKEN,
        X_V_S,
        X_V_S_VESTING,
        X_V_S_VESTING_PROXY,
        V_A_I_VAULT,
        V_A_I_VAULT_PROXY,
        V_R_T_VAULT,
        V_R_T_VAULT_PROXY,
        X_V_S_VAULT,
        X_V_S_VAULT_ERROR_REPORTER,
        X_V_S_VAULT_PROXY
    ]

if __name__ == "__main__":
    result = []
    for i in VENUS_ABIS.mapping:
        for obj in i:
            if obj.get('type') == "event":
                if "signature" in obj:
                    del obj['signature']
                result.append(obj)
    with open("events.json", 'w') as f:
        json.dump(result, f, indent=2)


from defi_services.abis.lending.compound.compound_v3.comet_abi import COMET_ABI
from defi_services.abis.lending.compound.compound_v3.comet_ext_abi import COMET_EXT_ABI
from defi_services.abis.lending.compound.compound_v3.reward_abi import REWARD_ABI
from defi_services.abis.token.erc20_abi import ERC20_ABI
from defi_services.constants.chain_constant import Chain
from defi_services.constants.entities.lending_constant import Lending
from defi_services.constants.time_constant import TimeConstants
from defi_services.constants.token_constant import Token
from defi_services.jobs.queriers.state_querier import StateQuerier
from defi_services.services.lending.compound_service import CompoundStateService
from defi_services.services.lending.lending_info.arbitrum.compound_v3_arbitrum import COMPOUND_V3_ARBITRUM
from defi_services.services.lending.lending_info.ethereum.compound_v3_eth import COMPOUND_V3_ETH
from defi_services.services.lending.lending_info.polygon.compound_v3_polygon import COMPOUND_V3_POLYGON


class CompoundV3Info:
    mapping = {
        Chain.ethereum: COMPOUND_V3_ETH,
        Chain.polygon: COMPOUND_V3_POLYGON,
        Chain.arbitrum: COMPOUND_V3_ARBITRUM
    }


class CompoundV3StateService(CompoundStateService):
    def __init__(self, state_service: StateQuerier, chain_id: str = "0x1"):
        super().__init__(state_service, chain_id)
        self.name = f"{chain_id}_{Lending.compound_v3}"
        self.pool_info = CompoundV3Info.mapping.get(chain_id)
        self.comet_abi = COMET_ABI
        self.comet_ext = COMET_EXT_ABI
        self.reward_abi = REWARD_ABI

    def get_service_info(self):
        info = {
            Lending.compound_v3: {
                "chain_id": self.chain_id,
                "type": "lending",
                "protocol_info": self.pool_info
            }
        }
        return info

    def get_dapp_asset_info(
            self,
            block_number: int = "latest",
            comets: list = None,
    ):
        result = {}
        w3 = self.state_service.get_w3()
        pools = [value.get('comet') for key, value in self.pool_info.get('reservesList').items()]
        if comets:
            pools += comets
        for pool in pools:
            contract = w3.eth.contract(address=w3.to_checksum_address(pool), abi=self.comet_abi)
            base_token = contract.functions.baseToken().call().lower()
            number_assets = contract.functions.numAssets().call()
            asset_data = {}
            for i in range(number_assets):
                data = contract.functions.getAssetInfo(i).call()
                asset = data[1].lower()
                asset_data[asset] = {
                    "priceFeed": data[2].lower(),
                    "loanToValue": data[4] / 10 ** 18,
                    "liquidationThreshold": data[5] / 10 ** 18
                }
            result[base_token] = {
                "comet": pool.lower(),
                "assets": asset_data
            }
        return result

    # PROTOCOL APY
    def get_apy_lending_pool_function_info(
            self,
            reserves_info: dict,
            block_number: int = "latest"
    ):
        rpc_calls = {}
        for token_address, reserve_info in reserves_info.items():
            if token_address != Token.native_token:
                query_id = f"decimals_{token_address}_{block_number}".lower()
                rpc_calls[query_id] = self.state_service.get_function_info(token_address, ERC20_ABI, "decimals", [], block_number)

            comet_address = reserve_info['comet']
            functions = [
                'decimals', 'totalBorrow', 'getUtilization', 'borrowKink', 'borrowPerSecondInterestRateBase',
                'borrowPerSecondInterestRateSlopeHigh', 'borrowPerSecondInterestRateSlopeLow',
                'supplyKink', 'supplyPerSecondInterestRateBase', 'supplyPerSecondInterestRateSlopeHigh',
                'supplyPerSecondInterestRateSlopeLow'
            ]
            for fn_name in functions:
                query_id = f"{fn_name}_{comet_address}_{block_number}".lower()
                rpc_calls[query_id] = self.get_comet_function_info(
                    comet=comet_address,
                    fn_name=fn_name,
                    block_number=block_number
                )

            for collateral_address in reserve_info['assets']:
                decimals_query_id = f"decimals_{collateral_address}_{block_number}".lower()
                rpc_calls[decimals_query_id] = self.state_service.get_function_info(collateral_address, ERC20_ABI, "decimals", [], block_number)

                supply_query_id = f"totalsCollateral_{comet_address}_{collateral_address}_{block_number}".lower()
                rpc_calls[supply_query_id] = self.get_comet_function_info(
                    comet=comet_address,
                    fn_name='totalsCollateral',
                    fn_paras=[collateral_address],
                    block_number=block_number
                )

        return rpc_calls

    def get_reserve_tokens_metadata(
            self,
            decoded_data: dict,
            reserves_info: dict,
            block_number: int = "latest"
    ):
        reserve_tokens_info = []
        for token_address, reserve_info in reserves_info.items():
            if token_address != Token.native_token:
                underlying_decimals_query_id = f"decimals_{token_address}_{block_number}".lower()
                underlying_decimals = decoded_data.get(underlying_decimals_query_id)
            else:
                underlying_decimals = Chain.native_decimals.get(self.chain_id, 18)

            comet_address = reserve_info['comet']
            comet_decimals_query_id = f"decimals_{comet_address}_{block_number}".lower()
            total_borrow_query_id = f"totalBorrow_{comet_address}_{block_number}".lower()
            utilization_query_id = f"getUtilization_{comet_address}_{block_number}".lower()

            info = {
                "token": comet_address,
                "token_decimals": decoded_data.get(comet_decimals_query_id),
                "utilization": decoded_data.get(utilization_query_id),
                'total_borrow': decoded_data.get(total_borrow_query_id),
                "underlying_decimals": underlying_decimals,
                "underlying": token_address
            }
            for lending_type in ['borrow', 'supply']:
                kink_query_id = f"{lending_type}Kink_{comet_address}_{block_number}".lower()
                base_rate_query_id = f"{lending_type}PerSecondInterestRateBase_{comet_address}_{block_number}".lower()
                jump_multiplier_query_id = f"{lending_type}PerSecondInterestRateSlopeHigh_{comet_address}_{block_number}".lower()
                multiplier_query_id = f"{lending_type}PerSecondInterestRateSlopeLow_{comet_address}_{block_number}".lower()

                info[f'{lending_type}_kink'] = decoded_data.get(kink_query_id)
                info[f'{lending_type}_base_rate'] = decoded_data.get(base_rate_query_id)
                info[f'{lending_type}_jump_multiplier'] = decoded_data.get(jump_multiplier_query_id)
                info[f'{lending_type}_multiplier'] = decoded_data.get(multiplier_query_id)

            assets = []
            for collateral_address in reserve_info['assets']:
                decimals_query_id = f"decimals_{collateral_address}_{block_number}".lower()
                supply_query_id = f"totalsCollateral_{comet_address}_{collateral_address}_{block_number}".lower()

                assets.append({
                    'underlying': collateral_address,
                    'total_supply': decoded_data.get(supply_query_id)[0],
                    'underlying_decimals': decoded_data.get(decimals_query_id)
                })
            info['assets'] = assets
            reserve_tokens_info.append(info)

        return reserve_tokens_info

    def calculate_apy_lending_pool_function_call(
            self,
            reserves_info: dict,
            decoded_data: dict,
            token_prices: dict,
            pool_token_price: float,
            pool_decimals: int = 18,
            block_number: int = "latest",
    ):
        reserve_tokens_info = self.get_reserve_tokens_metadata(decoded_data, reserves_info, block_number)

        data = {}
        for token_info in reserve_tokens_info:
            underlying_token = token_info['underlying']

            asset_info = self._calculate_interest_rates(
                token_info, pool_decimals=pool_decimals,
                apx_block_speed_in_seconds=0
            )

            assets = {
                underlying_token: {
                    'deposit_apy': 0,
                    'borrow_apy': asset_info['borrow_apy'],
                    'total_deposit': 0,
                    'total_borrow': asset_info['total_borrow'],
                    'is_base': True
                }
            }
            collaterals = asset_info.get('collaterals', {})
            for collateral_address, total_supply in collaterals.items():
                if collateral_address not in assets:
                    assets[collateral_address] = {
                        'deposit_apy': 0,
                        'borrow_apy': 0,
                        'total_deposit': 0,
                        'total_borrow': 0,
                        'is_base': False
                    }
                assets[collateral_address]['total_deposit'] += total_supply
                assets[collateral_address]['deposit_apy'] = asset_info['deposit_apy']

            comet = token_info['token']
            data[comet] = assets

        return data

    @classmethod
    def _calculate_interest_rates(
            cls, token_info: dict, pool_decimals: int, apx_block_speed_in_seconds: float):

        total_borrow = float(token_info["total_borrow"]) / 10 ** int(token_info["underlying_decimals"])
        collaterals = {}
        for asset_info in token_info['assets']:
            collateral_address = asset_info['underlying']
            total_supply = float(asset_info["total_supply"]) / 10 ** int(asset_info["underlying_decimals"])
            collaterals[collateral_address] = total_supply

        utilization = token_info['utilization'] / 10 ** 18
        interests = {}
        for lending_type in ['borrow', 'supply']:
            kink = token_info[f'{lending_type}_kink'] / 10 ** 18
            base_rate = TimeConstants.A_YEAR * token_info[f'{lending_type}_base_rate'] / 10 ** 18
            jump_multiplier = TimeConstants.A_YEAR * token_info[f'{lending_type}_jump_multiplier'] / 10 ** 18
            multiplier = TimeConstants.A_YEAR * token_info[f'{lending_type}_multiplier'] / 10 ** 18

            interest_rate = multiplier * min(utilization, kink) + jump_multiplier * max(0, utilization - kink) + base_rate
            interests[lending_type] = interest_rate

        return {
            'deposit_apy': interests['supply'],
            'borrow_apy': interests['borrow'],
            'total_deposit': None,
            'total_borrow': total_borrow,
            'collaterals': collaterals
        }

    # REWARDS BALANCE
    def get_rewards_balance_function_info(
            self,
            wallet: str,
            reserves_info: dict = None,
            block_number: int = "latest",
    ):
        result = {}
        reward_address = self.pool_info.get("rewardAddress")
        for token, value in self.pool_info.get("reservesList").items():
            comet = value.get("comet")
            rpc_call = self.state_service.get_function_info(
                reward_address, self.reward_abi, "getRewardOwed", [comet, wallet], block_number)
            get_reward_id = f"getRewardOwed_{reward_address}_{comet}_{wallet}_{block_number}".lower()
            result[get_reward_id] = rpc_call
        return result

    def calculate_rewards_balance(
            self, wallet: str, reserves_info: dict, decoded_data: dict, block_number: int = "latest"):
        reward_amount = 0
        reward_address = self.pool_info.get("rewardAddress")
        for token, value in self.pool_info.get("reservesList").items():
            comet = value.get("comet")
            get_reward_id = f"getRewardOwed_{reward_address}_{comet}_{wallet}_{block_number}".lower()
            reward_amount += decoded_data.get(get_reward_id)[1] / 10 ** 18
        reward_token = self.pool_info.get("rewardToken")
        result = {
            reward_token: {"amount": reward_amount}
        }
        return result

    # WALLET DEPOSIT BORROW BALANCE
    def get_wallet_deposit_borrow_balance_function_info(
            self,
            wallet: str,
            reserves_info: dict,
            block_number: int = "latest",
            health_factor: bool = False
    ):

        rpc_calls = {}
        for token, value in reserves_info.items():
            underlying = token
            if token == Token.native_token:
                underlying = Token.wrapped_token.get(self.chain_id)
            comet = value.get('comet')
            comet_ext = value.get('comet')
            assets = value.get('assets')
            decimals_key = f"decimals_{underlying}_{block_number}".lower()
            rpc_calls[decimals_key] = self.state_service.get_function_info(
                    underlying, ERC20_ABI, "decimals", [], block_number
                )
            borrow_key = f"borrowBalanceOf_{comet}_{underlying}_{wallet}_{block_number}".lower()
            rpc_calls[borrow_key] = self.state_service.get_function_info(
                comet, self.comet_abi, "borrowBalanceOf", [wallet], block_number)
            for asset in assets:
                balance_key = f"collateralBalanceOf_{comet_ext}_{asset}_{wallet}_{block_number}".lower()
                decimals_key = f"decimals_{asset}_{block_number}".lower()
                rpc_calls[balance_key] = self.state_service.get_function_info(
                    comet_ext, self.comet_ext, "collateralBalanceOf", [wallet, asset], block_number)
                rpc_calls[decimals_key] = self.state_service.get_function_info(
                    asset, ERC20_ABI, "decimals", [], block_number
                )

        return rpc_calls

    def calculate_wallet_deposit_borrow_balance(
            self,
            wallet: str,
            reserves_info: dict,
            decoded_data: dict,
            token_prices: dict = None,
            pool_decimals: int = 18,
            block_number: int = "latest",
            health_factor: bool = False
    ):
        result = {}
        total_borrow = 0
        total_collateral = 0
        for token, value in reserves_info.items():
            deposit_borrow = {}
            underlying = token
            if token == Token.native_token:
                underlying = Token.wrapped_token.get(self.chain_id)
            comet = value.get('comet')
            comet_ext = value.get('comet')
            assets = value.get('assets')
            get_decimals_id = f"decimals_{underlying}_{block_number}".lower()
            borrow_key = f"borrowBalanceOf_{comet}_{underlying}_{wallet}_{block_number}".lower()
            decimals = decoded_data[get_decimals_id]
            borrow_amount = decoded_data[borrow_key] / 10 ** decimals
            for asset, asset_info in assets.items():
                balance_key = f"collateralBalanceOf_{comet_ext}_{asset}_{wallet}_{block_number}".lower()
                decimals_key = f"decimals_{asset}_{block_number}".lower()
                decimals_token = decoded_data[decimals_key]
                deposit_amount = decoded_data[balance_key] / 10 ** decimals_token
                deposit_borrow[asset] = {
                        "borrow_amount": 0,
                        "deposit_amount": deposit_amount,
                        "is_collateral": True
                    }
                if token_prices:
                    deposit_amount_in_usd = deposit_amount * token_prices.get(asset)
                    total_collateral += deposit_amount_in_usd * asset_info.get("liquidationThreshold")

            if underlying not in deposit_borrow:
                deposit_borrow[underlying] = {
                    "borrow_amount": borrow_amount,
                    "deposit_amount": 0,
                    "is_collateral": False
                }
            else:
                deposit_borrow[underlying]["borrow_amount"] = borrow_amount

            if token_prices:
                borrow_amount_in_usd = token_prices.get(underlying)
                total_borrow += borrow_amount_in_usd

            result[comet] = deposit_borrow

        if health_factor:
            if total_collateral and total_borrow:
                result['health_factor'] = total_collateral / total_borrow
            elif total_collateral:
                result['health_factor'] = 100
            else:
                result['health_factor'] = 0

        return result

    def get_comet_function_info(self, comet: str, fn_name: str, fn_paras: list = None, block_number: int = "latest"):
        return self.state_service.get_function_info(
            comet, self.comet_abi, fn_name, fn_paras, block_number
        )



import json

ONYX_LENS_ABI = json.loads('''
[
  {
    "constant": true,
    "inputs": [],
    "name": "BLOCKS_PER_DAY",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "internalType": "contract ComptrollerLensInterface",
        "name": "comptroller",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "getAccountLimits",
    "outputs": [
      {
        "components": [
          {
            "internalType": "contract OToken[]",
            "name": "markets",
            "type": "address[]"
          },
          {
            "internalType": "uint256",
            "name": "liquidity",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "shortfall",
            "type": "uint256"
          }
        ],
        "internalType": "struct OnyxLens.AccountLimits",
        "name": "",
        "type": "tuple"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "internalType": "address payable",
        "name": "account",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "comptrollerAddress",
        "type": "address"
      }
    ],
    "name": "getDailyXCN",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "internalType": "contract GovernorBravoInterface",
        "name": "governor",
        "type": "address"
      },
      {
        "internalType": "uint256[]",
        "name": "proposalIds",
        "type": "uint256[]"
      }
    ],
    "name": "getGovBravoProposals",
    "outputs": [
      {
        "components": [
          {
            "internalType": "uint256",
            "name": "proposalId",
            "type": "uint256"
          },
          {
            "internalType": "address",
            "name": "proposer",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "eta",
            "type": "uint256"
          },
          {
            "internalType": "address[]",
            "name": "targets",
            "type": "address[]"
          },
          {
            "internalType": "uint256[]",
            "name": "values",
            "type": "uint256[]"
          },
          {
            "internalType": "string[]",
            "name": "signatures",
            "type": "string[]"
          },
          {
            "internalType": "bytes[]",
            "name": "calldatas",
            "type": "bytes[]"
          },
          {
            "internalType": "uint256",
            "name": "startBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "endBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "forVotes",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "againstVotes",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "abstainVotes",
            "type": "uint256"
          },
          {
            "internalType": "bool",
            "name": "canceled",
            "type": "bool"
          },
          {
            "internalType": "bool",
            "name": "executed",
            "type": "bool"
          }
        ],
        "internalType": "struct OnyxLens.GovBravoProposal[]",
        "name": "",
        "type": "tuple[]"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "internalType": "contract GovernorBravoInterface",
        "name": "governor",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "voter",
        "type": "address"
      },
      {
        "internalType": "uint256[]",
        "name": "proposalIds",
        "type": "uint256[]"
      }
    ],
    "name": "getGovBravoReceipts",
    "outputs": [
      {
        "components": [
          {
            "internalType": "uint256",
            "name": "proposalId",
            "type": "uint256"
          },
          {
            "internalType": "bool",
            "name": "hasVoted",
            "type": "bool"
          },
          {
            "internalType": "uint8",
            "name": "support",
            "type": "uint8"
          },
          {
            "internalType": "uint96",
            "name": "votes",
            "type": "uint96"
          }
        ],
        "internalType": "struct OnyxLens.GovBravoReceipt[]",
        "name": "",
        "type": "tuple[]"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "internalType": "contract Xcn",
        "name": "xcn",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "getXcnBalanceMetadata",
    "outputs": [
      {
        "components": [
          {
            "internalType": "uint256",
            "name": "balance",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "votes",
            "type": "uint256"
          },
          {
            "internalType": "address",
            "name": "delegate",
            "type": "address"
          }
        ],
        "internalType": "struct OnyxLens.XcnBalanceMetadata",
        "name": "",
        "type": "tuple"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "internalType": "contract Xcn",
        "name": "xcn",
        "type": "address"
      },
      {
        "internalType": "contract ComptrollerLensInterface",
        "name": "comptroller",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "getXcnBalanceMetadataExt",
    "outputs": [
      {
        "components": [
          {
            "internalType": "uint256",
            "name": "balance",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "votes",
            "type": "uint256"
          },
          {
            "internalType": "address",
            "name": "delegate",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "allocated",
            "type": "uint256"
          }
        ],
        "internalType": "struct OnyxLens.XcnBalanceMetadataExt",
        "name": "",
        "type": "tuple"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "internalType": "contract Xcn",
        "name": "xcn",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      },
      {
        "internalType": "uint32[]",
        "name": "blockNumbers",
        "type": "uint32[]"
      }
    ],
    "name": "getXcnVotes",
    "outputs": [
      {
        "components": [
          {
            "internalType": "uint256",
            "name": "blockNumber",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "votes",
            "type": "uint256"
          }
        ],
        "internalType": "struct OnyxLens.XcnVotes[]",
        "name": "",
        "type": "tuple[]"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "internalType": "contract OToken",
        "name": "oToken",
        "type": "address"
      },
      {
        "internalType": "address payable",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "oTokenBalances",
    "outputs": [
      {
        "components": [
          {
            "internalType": "address",
            "name": "oToken",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "balanceOf",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "borrowBalanceCurrent",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "balanceOfUnderlying",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "tokenBalance",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "tokenAllowance",
            "type": "uint256"
          }
        ],
        "internalType": "struct OnyxLens.OTokenBalances",
        "name": "",
        "type": "tuple"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "internalType": "contract OToken[]",
        "name": "oTokens",
        "type": "address[]"
      },
      {
        "internalType": "address payable",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "oTokenBalancesAll",
    "outputs": [
      {
        "components": [
          {
            "internalType": "address",
            "name": "oToken",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "balanceOf",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "borrowBalanceCurrent",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "balanceOfUnderlying",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "tokenBalance",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "tokenAllowance",
            "type": "uint256"
          }
        ],
        "internalType": "struct OnyxLens.OTokenBalances[]",
        "name": "",
        "type": "tuple[]"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "internalType": "contract OToken",
        "name": "oToken",
        "type": "address"
      }
    ],
    "name": "oTokenMetadata",
    "outputs": [
      {
        "components": [
          {
            "internalType": "address",
            "name": "oToken",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "exchangeRateCurrent",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "supplyRatePerBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "borrowRatePerBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "reserveFactorMantissa",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalBorrows",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalReserves",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalSupply",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalCash",
            "type": "uint256"
          },
          {
            "internalType": "bool",
            "name": "isListed",
            "type": "bool"
          },
          {
            "internalType": "uint256",
            "name": "collateralFactorMantissa",
            "type": "uint256"
          },
          {
            "internalType": "address",
            "name": "underlyingAssetAddress",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "oTokenDecimals",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "underlyingDecimals",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "xcnSupplySpeed",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "xcnBorrowSpeed",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "borrowCap",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "dailySupplyXcn",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "dailyBorrowXcn",
            "type": "uint256"
          }
        ],
        "internalType": "struct OnyxLens.OTokenMetadata",
        "name": "",
        "type": "tuple"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "internalType": "contract OToken[]",
        "name": "oTokens",
        "type": "address[]"
      }
    ],
    "name": "oTokenMetadataAll",
    "outputs": [
      {
        "components": [
          {
            "internalType": "address",
            "name": "oToken",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "exchangeRateCurrent",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "supplyRatePerBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "borrowRatePerBlock",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "reserveFactorMantissa",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalBorrows",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalReserves",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalSupply",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "totalCash",
            "type": "uint256"
          },
          {
            "internalType": "bool",
            "name": "isListed",
            "type": "bool"
          },
          {
            "internalType": "uint256",
            "name": "collateralFactorMantissa",
            "type": "uint256"
          },
          {
            "internalType": "address",
            "name": "underlyingAssetAddress",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "oTokenDecimals",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "underlyingDecimals",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "xcnSupplySpeed",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "xcnBorrowSpeed",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "borrowCap",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "dailySupplyXcn",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "dailyBorrowXcn",
            "type": "uint256"
          }
        ],
        "internalType": "struct OnyxLens.OTokenMetadata[]",
        "name": "",
        "type": "tuple[]"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "internalType": "contract OToken",
        "name": "oToken",
        "type": "address"
      }
    ],
    "name": "oTokenUnderlyingPrice",
    "outputs": [
      {
        "components": [
          {
            "internalType": "address",
            "name": "oToken",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "underlyingPrice",
            "type": "uint256"
          }
        ],
        "internalType": "struct OnyxLens.OTokenUnderlyingPrice",
        "name": "",
        "type": "tuple"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "internalType": "contract OToken[]",
        "name": "oTokens",
        "type": "address[]"
      }
    ],
    "name": "oTokenUnderlyingPriceAll",
    "outputs": [
      {
        "components": [
          {
            "internalType": "address",
            "name": "oToken",
            "type": "address"
          },
          {
            "internalType": "uint256",
            "name": "underlyingPrice",
            "type": "uint256"
          }
        ],
        "internalType": "struct OnyxLens.OTokenUnderlyingPrice[]",
        "name": "",
        "type": "tuple[]"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "internalType": "address",
        "name": "holder",
        "type": "address"
      },
      {
        "internalType": "contract ComptrollerLensInterface",
        "name": "comptroller",
        "type": "address"
      }
    ],
    "name": "pendingXcn",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  }
]
''')


import json

LIDO_REMOVE_ALLOWED_RECIPIENT_ABI = json.loads('''
[
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_trustedCaller",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_allowedRecipientsRegistry",
        "type": "address"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "inputs": [],
    "name": "allowedRecipientsRegistry",
    "outputs": [
      {
        "internalType": "contract AllowedRecipientsRegistry",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_creator",
        "type": "address"
      },
      {
        "internalType": "bytes",
        "name": "_evmScriptCallData",
        "type": "bytes"
      }
    ],
    "name": "createEVMScript",
    "outputs": [
      {
        "internalType": "bytes",
        "name": "",
        "type": "bytes"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bytes",
        "name": "_evmScriptCallData",
        "type": "bytes"
      }
    ],
    "name": "decodeEVMScriptCallData",
    "outputs": [
      {
        "internalType": "address",
        "name": "recipientAddress",
        "type": "address"
      }
    ],
    "stateMutability": "pure",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "trustedCaller",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  }
]
''')


from defi_services.constants.chain_constant import Chain
from defi_services.constants.entities.lending_constant import Lending


class Token:
    native_token = '0x0000000000000000000000000000000000000000'
    wrapped_ethereum = '0xc02aaa39b223fe8d0a0e5c4f27ead9083c756cc2'
    wrapped_bsc = '0xbb4cdb9cbd36b01bd1cbaebf2de08d9173bc095c'
    wrapped_ftm = '0x21be370d5312f44cb42ce377bc9b8a0cef1a4c83'
    wrapped_avax = "0xb31f66aa3c1e785363f0875a1b74e27b85fd66c7"
    wrapped_eth_optimism = "0x4200000000000000000000000000000000000006"
    wrapped_eth_arbitrum = "0x82af49447d8a07e3bd95bd0d56f35241523fbab1"
    wrapped_matic = '0x0000000000000000000000000000000000001010'
    wrapped_sol = 'So11111111111111111111111111111111111111112'
    wrapped_tron = '0x891cdb91d149f23b1a45d9c5ca78a88d0cb44c18'
    wrapped_cro = '0x5c7f8a570d578ed84e63fdfa7b1ee72deae1ae23'

    wrapped_token = {
        Chain.solana: wrapped_sol,
        Chain.ethereum: wrapped_ethereum,
        Chain.bsc: wrapped_bsc,
        Chain.fantom: wrapped_ftm,
        Chain.avalanche: wrapped_avax,
        Chain.optimism: wrapped_eth_optimism,
        Chain.arbitrum: wrapped_eth_arbitrum,
        Chain.polygon: wrapped_matic,
        Chain.tron: wrapped_tron,
        Chain.cronos: wrapped_cro,
        Chain.base: '0x4200000000000000000000000000000000000006',
        Chain.kava: '0xc86c7c0efbd6a49b35e8714c5f59d99de09a225b',
        Chain.gnosis: '0xe91d153e0b41518a2ce8dd3d7944fa863463a97d',
        Chain.klaytn: '0x19aac5f612f524b754ca7e7c41cbfa2e981a4432',
        Chain.mantle: '0x78c1b0c915c4faa5fffa6cabf0219da63d7f4cb8',
        Chain.celo: '0x3ad443d769a07f287806874f8e5405ce3ac902b9',
        Chain.moonbeam: '0xacc15dc74880c9944775448304b263d191c6077f',
        Chain.manta: '0x0dc808adce2099a9f62aa87d9670745aba741746',
        Chain.pulse: '0xa1077a294dde1b09bb078844df40758a5d0f9a27',
        Chain.rootstock: '0x967f8799af07df1534d48a95a5c9febe92c53ae0',
        Chain.astar: '0xaeaaf0e2c81af264101b9129c00f4440ccf0f720',
        Chain.metis: '0x75cb093e4d61d2a2e65d8e0bbb01de8d89b53481',
        Chain.canto: '0x826551890dc65655a0aceca109ab11abdbd7a07b',
        Chain.heco: '0x5545153ccfca01fbd7dd11c0b23ba694d9509a6f',
        Chain.linea: '0xe5d7c2a44ffddf6b295a15c148167daaaf5cf34f',
        Chain.okc: '0x8f8526dbfd6e38e3d8307702ca8469bae6c56c15',
        Chain.aurora: '0xc9bdeed33cd01541e1eed10f90519d2c06fe3feb',
        Chain.moonriver: '0x98878b06940ae243284ca214f92bb71a2b032b8a',
        Chain.blast: '0x4300000000000000000000000000000000000004',
        Chain.oasis_sapphire: '0x8bc2b030b299964eefb5e1e0b36991352e56d2d3'
    }


class ContractAddresses:
    WBNB = "0xbb4cdb9cbd36b01bd1cbaebf2de08d9173bc095c"
    BNB = "0x0000000000000000000000000000000000000000"
    TRAVA_BNB_LP = "0x865c77d4ff6383e06c58350a2cfb95cca2c0f056"
    R_TRAVA = "0x170772a06affc0d375ce90ef59c8ec04c7ebf5d2"
    VE_TRAVA_VAULT = "0xedf6a93772eecfb1300f61f6c1ae9680c33996a2"
    GENERAL_TRAVA_VAULT = "63daeb1c97e098480b16c7410dc4b9c0d4c401d6ff4735324edfbeaa"
    PANCAKE_FARM = "0x73feaa1ee314f8c655e354234017be2193c9e24e"
    # PANCAKE_FARM = "0xa5f8c5dbd5f286960b9d90548680ae5ebff07652"
    TRAVA_BSC_LENDING_POOL = "0x75de5f7c91a89c16714017c7443eca20c7a8c295"
    VALAS_BSC_LENDING_POOL = "0xe29a55a6aeff5c8b1beede5bcf2f0cb3af8f91f5"
    ALPACA_BSC_LENDING_POOL = "0xa625ab01b08ce023b2a342dbb12a16f2c8489a8f"
    DOGE = "0xba2ae424d960c26247dd6c32edc70b295c744c43"
    MAIN_STAKING_CONTRACT_PANCAKE_FARM = "0x73feaa1ee314f8c655e354234017be2193c9e24e"
    # MAIN_STAKING_CONTRACT_PANCAKE_FARM = "0xa5f8c5dbd5f286960b9d90548680ae5ebff07652"
    LUNA = "0xb91A659E88B51474767CD97EF3196A3e7cEDD2c8"
    UST = "0x78366446547D062f45b4C0f320cDaa6d710D87bb"

    BISWAP_FARM = "0xdbc1a13490deef9c3c12b44fe77b503c1b061739"
    PANCAKE_FARM_V2 = "0xa5f8C5Dbd5F286960b9d90548680aE5ebFf07652"
    SUSHISWAP_FARM = "0xc2edad668740f1aa35e4d8f227fb8e17dca888cd"
    SUSHISWAP_FARM_V2 = '0xef0881ec094552b2e128cf945ef17a6752b4ec5d'
    MINICHEF_ARBITRUM_NOVA = '0xC09756432dAD2FF50B2D40618f7B04546DD20043'
    MINICHEF_ARBITRUM_ONE = '0xF4d73326C13a4Fc5FD7A064217e12780e9Bd62c3'
    MINICHEF_AVALANCHE = '0xe11252176CEDd4a493Aec9767192C06A04A6B04F'
    MINICHEF_BSC = '0x5219C5E32b9FFf87F29d5A833832c29134464aaa'
    MINICHEF_FANTOM = "0xf731202A3cf7EfA9368C2d7bD613926f7A144dB5"
    MINICHEF_OPTIMISM = '0xB25157bF349295a7Cd31D1751973f426182070D6'
    MINICHEF_POLYGON = '0x0769fd68dFb93167989C6f7254cd0D766Fb2841F'
    SUSHISWAP_MINICHEF = {
        Chain.bsc: MINICHEF_BSC,
        Chain.fantom: MINICHEF_FANTOM,
        Chain.polygon: MINICHEF_POLYGON,
        Chain.optimism: MINICHEF_OPTIMISM,
        Chain.avalanche: MINICHEF_AVALANCHE,
    }


class ProtocolNFT:
    nft = {
        "0x60e4d786628fea6478f785a6d7e704777c86a7c6": Lending.onyx,
        "0xb7f7f6c52f2e2fdb1963eab30438024864c313f6": Lending.onyx,
        "0xbc4ca0eda7647a8ab7c2061c2e118a18a936f13d": Lending.onyx
    }



def get_fees(fee_growth_global_0, fee_growth_global_1, fee_growth_0_low, fee_growth_0_hi, fee_growth_inside_0,
             fee_growth_1_low, fee_growth_1_hi, fee_growth_inside_1, liquidity, decimals0, decimals1, tick_lower,
             tick_upper, tick_current):
    if tick_current >= tick_upper:
        tick_upper_fee_growth_above_0 = sub_in_256(fee_growth_global_0, fee_growth_0_hi)
        tick_upper_fee_growth_above_1 = sub_in_256(fee_growth_global_1, fee_growth_1_hi)
    else:
        tick_upper_fee_growth_above_0 = fee_growth_0_hi
        tick_upper_fee_growth_above_1 = fee_growth_1_hi
    if tick_current >= tick_lower:
        tick_lower_fee_growth_below_0 = fee_growth_0_low
        tick_lower_fee_growth_below_1 = fee_growth_1_low
    else:
        tick_lower_fee_growth_below_0 = sub_in_256(fee_growth_global_0, fee_growth_0_low)
        tick_lower_fee_growth_below_1 = sub_in_256(fee_growth_global_1, fee_growth_1_low)

    fr_t1_0 = sub_in_256(sub_in_256(fee_growth_global_0, tick_lower_fee_growth_below_0),
                         tick_upper_fee_growth_above_0)
    fr_t1_1 = sub_in_256(sub_in_256(fee_growth_global_1, tick_lower_fee_growth_below_1),
                         tick_upper_fee_growth_above_1)
    uncollect_fee_0 = liquidity * sub_in_256(fr_t1_0, fee_growth_inside_0) / 2 ** 128
    uncollect_fee_1 = liquidity * sub_in_256(fr_t1_1, fee_growth_inside_1) / 2 ** 128
    return uncollect_fee_0 / 10 ** decimals0, uncollect_fee_1 / 10 ** decimals1


def sub_in_256(a: int, b: int) -> int:
    """
    Performs subtraction on two 256-bit unsigned integers, handling overflow.
    """

    MAX_UINT256 = (1 << 256) - 1
    result = (a - b) % (MAX_UINT256 + 1)  # Handle overflow using modulo
    return result



import os

import dotenv

dotenv.load_dotenv()

POLYGON_AAVE_ADDRESS = '0x8dff5e27ea6b7ac08ebfdf9eb090f32ee9a30fcf'
ETHEREUM_AAVE_ADDRESS = '0x7d2768de32b0b80b7a3454c06bdac94a69ddc7a9'

BSC_TRAVA_ADDRESS = '0x75de5f7c91a89c16714017c7443eca20c7a8c295'
ETH_TRAVA_ADDRESS = '0xd61afaaa8a69ba541bc4db9c9b40d4142b43b9a4'
FTM_TRAVA_ADDRESS = '0xd98bb590bdfabf18c164056c185fbb6be5ee643f'

BSC_VALAS_ADDRESS = '0xe29a55a6aeff5c8b1beede5bcf2f0cb3af8f91f5'

BSC_VENUS_ADDRESS = '0xfd36e2c2a6789db23113685031d7f16329158384'
BSC_CREAM_ADDRESS = '0x589de0f0ccf905477646599bb3e5c622c84cc0ba'

FTM_GEIST_ADDRESS = '0x9fad24f572045c7869117160a571b2e50b10d068'

ETH_COMPOUND_ADDRESS = '0x3d9819210a31b4961b30ef54be2aed79b9c9cd3b'
ETH_CREAM_ADDRESS = ''


class Chains:
    bsc = '0x38'
    ethereum = '0x1'
    fantom = '0xfa'
    polygon = '0x89'
    arbitrum = '0xa4b1'
    optimism = '0xa'
    avalanche = '0xa86a'
    tron = '0x2b6653dc'
    cronos = '0x19'
    solana = 'solana'
    polkadot = 'polkadot'
    bitcoin = 'bitcoin'
    cosmos = 'cosmos'
    base = '0x2105'
    kava = '0x8ae'
    gnosis = '0x64'
    klaytn = '0x2019'
    mantle = '0x1388'
    celo = '0xa4ec'
    moonbeam = '0x504'
    manta = '0xa9'
    pulse = '0x171'
    rootstock = '0x1e'
    astar = '0x250'
    metis = '0x440'
    canto = '0x1e14'
    heco = '0x80'
    linea = '0xe708'
    okc = '0x42'
    aurora = '0x4e454152'
    moonriver = '0x505'
    oasis_sapphire = '0x5afe'
    oasis_sapphire_testnet = '0x5aff'
    blast = '0xee'
    oraichain = 'orai'
    ton = 'ton'

    none_wrapped_token = [arbitrum, fantom, optimism]
    all = [
        bsc, ethereum, fantom, polygon, arbitrum, optimism,
        avalanche, tron, cronos, solana, polkadot, bitcoin, cosmos,
        oraichain, ton
    ]

    mapping = {
        'bsc': bsc,
        'ethereum': ethereum,
        'fantom': fantom,
        'polygon': polygon,
        'arbitrum': arbitrum,
        'optimism': optimism,
        'avalanche': avalanche,
        'tron': tron,
        'cronos': cronos,
        'solana': solana,
        'polkadot': polkadot,
        'oasis_sapphire': oasis_sapphire,
        'oasis_sapphire_testnet': oasis_sapphire_testnet,
        'ton': ton
    }

    names = {
        bsc: 'bsc',
        ethereum: "ethereum",
        fantom: 'fantom',
        polygon: 'polygon',
        arbitrum: 'arbitrum',
        optimism: 'optimism',
        avalanche: 'avalanche',
        tron: 'tron',
        cronos: 'cronos',
        solana: 'solana',
        polkadot: 'polkadot',
        oasis_sapphire: 'oasis_sapphire',
        oasis_sapphire_testnet: 'oasis_sapphire_testnet',
        ton: 'ton'
    }

    abi_mapping = {
        bsc: 'bep20_abi',
        ethereum: 'erc20_abi',
        fantom: 'erc20_abi',
        polygon: 'erc20_abi',
        arbitrum: 'erc20_abi',
        optimism: 'erc20_abi',
        avalanche: 'erc20_abi',
        tron: 'trc20_abi',
        cronos: 'crc20_abi'
    }

    block_time = {
        bsc: 3,
        ethereum: 12,
        fantom: 1,
        polygon: 2,
        arbitrum: 0.3,
        avalanche: 2,
        ton: 3
    }
    wrapped_native_token = {
        bsc: "0xbb4cdb9cbd36b01bd1cbaebf2de08d9173bc095c",
        ethereum: "0xc02aaa39b223fe8d0a0e5c4f27ead9083c756cc2",
        polygon: "0x0d500b1d8e8ef31e21c99d1db9a6444d3adf1270",
        fantom: "0x21be370d5312f44cb42ce377bc9b8a0cef1a4c83",
        arbitrum: "0x82af49447d8a07e3bd95bd0d56f35241523fbab1",
        optimism: "0x4200000000000000000000000000000000000006",
        avalanche: "0xb31f66aa3c1e785363f0875a1b74e27b85fd66c7",
        tron: "0x891cdb91d149f23b1a45d9c5ca78a88d0cb44c18",
        cronos: "0x5c7f8a570d578ed84e63fdfa7b1ee72deae1ae23",
    }

    evm = {
        ethereum: True, bsc: True, polygon: True, fantom: True,
        arbitrum: True, optimism: True, avalanche: True, tron: True, cronos: True,
        base: True, kava: True, gnosis: True, klaytn: True, mantle: True, celo: True, moonbeam: True, manta: True,
        pulse: True, rootstock: True, astar: True, metis: True, canto: True, heco: True, linea: True, okc: True,
        aurora: True, moonriver: True, oasis_sapphire: True, blast: True
    }

    def get_abi_name(self, chain_id):
        """Map chain_id with the corresponding abi key
        e.g.: '0x38' -> 'bep20_abi' or '0x1' -> 'erc20_abi'
        """
        return self.abi_mapping.get(chain_id, '')


class ChainsAnkr:
    """Chain name for Ankr API"""
    bsc = '0x38'
    eth = '0x1'
    fantom = '0xfa'
    polygon = '0x89'

    reversed_mapping = {
        bsc: 'bsc',
        eth: 'eth',
        fantom: 'fantom',
        polygon: 'polygon',
    }

    def get_chain_name(self, chain_id):
        return self.reversed_mapping.get(chain_id)


class Scans:
    mapping = {
        'bscscan': Chains.bsc,
        'etherscan': Chains.ethereum,
        'ftmscan': Chains.fantom,
        'polygonscan': Chains.polygon,
        'arbiscan': Chains.arbitrum,
        'optimistic': Chains.optimism,
        'snowtrace': Chains.avalanche,
        'cronoscan': Chains.cronos
    }

    api_keys = {
        Chains.bsc: 'GKRVTEQHSGWX335P21994DKNSTJVDYYXJ1',
        Chains.ethereum: 'YQVXQAJAYNS7AC48ZXS1NT45RKYRPY7P7E',
        Chains.fantom: 'Y85XY5DDDN4Z7C7Z4ZX8N62M5WET7D8NYC',
        Chains.polygon: 'EYADI5Q4VK4VF99KFEJURMVE9CNE54YEYW',
        Chains.arbitrum: 'BU5VN8872FE119RDX898MS5EZHE11FPZSC',
        Chains.optimism: '1JW9PDV2HZZI3WRFQRVGWUDD2T1RV8GWX3',
        Chains.avalanche: 'JWNM6Y5UA1499PWVA2T55MEH3TZDGAKJWJ',
        Chains.cronos: 'F5J6423D42PM21CDP61SUD6H7DE4FXH2DE'
    }

    api_bases = {
        Chains.bsc: 'https://api.bscscan.com/api',
        Chains.ethereum: 'https://api.etherscan.io/api',
        Chains.fantom: 'https://api.ftmscan.com/api',
        Chains.polygon: 'https://api.polygonscan.com/api',
        Chains.arbitrum: 'https://api.arbiscan.io/api',
        Chains.optimism: 'https://api-optimistic.etherscan.io/api',
        Chains.avalanche: 'https://api.snowtrace.io/api',
        Chains.cronos: 'https://cronoscan.com/api',
        Chains.ton: 'https://toncenter.com/api/'
    }

    scan_base_urls = {
        Chains.ethereum: 'https://etherscan.io/',
        Chains.bsc: 'https://bscscan.com/',
        Chains.fantom: 'https://ftmscan.com/',
        Chains.polygon: 'https://polygonscan.com/',
        Chains.arbitrum: 'https://arbiscan.io/',
        Chains.optimism: 'https://optimistic.etherscan.io/',
        Chains.avalanche: 'https://snowtrace.io/',
        Chains.cronos: 'https://cronoscan.com/',
        Chains.ton: 'https://toncenter.com/api/'
    }

    all_base_urls = {
        Chains.bsc: [
            'https://bscscan.com/token',
            'https://bscscan-com.translate.goog/token'
        ],
        Chains.polygon: [
            'https://polygonscan.com/token',
            'https://polygonscan-com.translate.goog/token'
        ],
        Chains.ethereum: [
            'https://etherscan.io/token',
            'https://etherscan-io.translate.goog/token',
        ],
        Chains.fantom: [
            'https://ftmscan.com/token',
            'https://ftmscan-com.translate.goog/token'
        ],
        Chains.arbitrum: [
            # 'https://arbiscan-io.translate.goog/token',
            'https://arbiscan.io/token'
        ],
        Chains.optimism: [
            # 'https://optimistic-etherscan-io.translate.goog/token',
            'https://optimistic.etherscan.io/token'
        ],
        # Chains.avalanche: [
        #     'https://snowtrace.io/token'
        # ],
        Chains.cronos: ['https://cronoscan.com/token']
    }

    gg_translate_suffix = '?_x_tr_sl=vi&_x_tr_tl=en&_x_tr_hl=en&_x_tr_pto=wapp'


class Networks:
    bsc = 'bsc'
    ethereum = 'ethereum'
    fantom = 'fantom'
    polygon = 'polygon'
    arbitrum = 'arbitrum'
    optimism = 'optimism'
    avalanche = 'avalanche'
    tron = 'tron'
    cronos = 'cronos'
    solana = 'solana'
    polkadot = 'polkadot'
    oasis_sapphire = 'oasis_sapphire'
    oasis_sapphire_testnet = 'oasis_sapphire_testnet'
    cosmos = 'cosmos'
    oraichain = 'oraichain'
    ton = 'ton'

    providers = {
        bsc: os.getenv('BSC_PROVIDER_URI', 'https://bsc-dataseed1.binance.org/'),
        ethereum: os.getenv('ETHEREUM_PROVIDER_URI', 'https://rpc.ankr.com/eth'),
        fantom: os.getenv('FANTOM_PROVIDER_URI', 'https://rpc.ftm.tools/'),
        polygon: os.getenv('POLYGON_PROVIDER_URI', 'https://polygon-rpc.com'),
        arbitrum: os.getenv('ARBITRUM_PROVIDER_URI', 'https://endpoints.omniatech.io/v1/arbitrum/one/public'),
        optimism: os.getenv('OPTIMISM_PROVIDER_URI', 'https://rpc.ankr.com/optimism'),
        avalanche: os.getenv('AVALANCHE_PROVIDER_URI', 'https://rpc.ankr.com/avalanche'),
        tron: os.getenv('TRON_PROVIDER_URI', 'https://rpc.ankr.com/tron_jsonrpc'),
        cronos: os.getenv('CRONOS_PROVIDER_URI', 'https://evm.cronos.org/'),
        solana: os.getenv('SOLANA_PROVIDER_URI', 'https://crimson-multi-putty.solana-mainnet.quiknode.pro/997174ce6ab5cc9d42cb037e931d18ae1a98346a/'),
        polkadot: os.getenv('POLKADOT_PROVIDER_URI', 'https://late-yolo-diagram.dot-mainnet.quiknode.pro/51a1aaf2372854dfd211fca3ab375e5451222be4/'),
        oasis_sapphire: os.getenv('OASIS_SAPPHIRE_PROVIDER_URI'),
        oasis_sapphire_testnet: os.getenv('OASIS_SAPPHIRE_PROVIDER_URI'),
        ton: os.getenv('TON_PROVIDER_URI')
    }

    archive_node = {
        bsc: os.getenv('BSC_PROVIDER_ARCHIVE_URI', 'https://rpc.ankr.com/bsc'),
        ethereum: os.getenv('ETHEREUM_PROVIDER_ARCHIVE_URI', 'https://rpc.ankr.com/eth'),
        fantom: os.getenv('FANTOM_PROVIDER_ARCHIVE_URI', 'https://rpc.ankr.com/fantom'),
        polygon: os.getenv('POLYGON_PROVIDER_ARCHIVE_URI', 'https://rpc.ankr.com/polygon'),
        arbitrum: os.getenv('ARBITRUM_PROVIDER_ARCHIVE_URI', 'https://rpc.ankr.com/arbitrum'),
        optimism: os.getenv('OPTIMISM_PROVIDER_ARCHIVE_URI', 'https://rpc.ankr.com/optimism'),
        avalanche: os.getenv('AVALANCHE_PROVIDER_ARCHIVE_URI', 'https://rpc.ankr.com/avalanche'),
        tron: os.getenv('TRON_PROVIDER_ARCHIVE_URI', 'https://rpc.ankr.com/tron_jsonrpc'),
        cronos: os.getenv('CRONOS_PROVIDER_ARCHIVE_URI'),
        solana: os.getenv('SOLANA_PROVIDER_ARCHIVE_URI'),
        polkadot: os.getenv('POLKADOT_PROVIDER_ARCHIVE_URI'),
        oasis_sapphire: os.getenv('OASIS_SAPPHIRE_PROVIDER_ARCHIVE_URI'),
        oasis_sapphire_testnet: os.getenv('OASIS_SAPPHIRE_PROVIDER_ARCHIVE_URI'),
        ton: os.getenv('TON_PROVIDER_ARCHIVE_URI', 'https://toncenter.com/api/')
    }


class DefiLlama:
    chains = {
        'Binance': Chains.bsc,
        'Ethereum': Chains.ethereum,
        'Fantom': Chains.fantom,
        'Polygon': Chains.polygon,
        'Arbitrum': Chains.arbitrum,
        'Optimism': Chains.optimism,
        'Avalanche': Chains.avalanche,
        'Tron': Chains.tron,
        'Cronos': Chains.cronos,
        'Solana': Chains.solana,
        'Polkadot': Chains.polkadot,
        'Base': Chains.base,
        'Kava': Chains.kava,
        'xDai': Chains.gnosis,
        'Klaytn': Chains.klaytn,
        'Mantle': Chains.mantle,
        'Celo': Chains.celo,
        'Moonbeam': Chains.moonbeam,
        'Manta': Chains.manta,
        'Pulse': Chains.pulse,
        'RSK': Chains.rootstock,
        'Astar': Chains.astar,
        'Metis': Chains.metis,
        'Canto': Chains.canto,
        'Heco': Chains.heco,
        'Linea': Chains.linea,
        'OKExChain': Chains.okc,
        'Aurora': Chains.aurora,
        'Moonriver': Chains.moonriver,
        'Oasis Sapphire': Chains.oasis_sapphire,
        'Blast': Chains.blast,
        'Oraichain': Chains.oraichain,
        'Ton': Chains.ton
    }


class Opensea:
    chains = {
        Chains.bsc: 'BSC',
        Chains.ethereum: 'ETHEREUM',
        Chains.polygon: 'MATIC',
        Chains.fantom: 'FANTOM',
        Chains.arbitrum: 'ARBITRUM',
        Chains.optimism: 'OPTIMISM',
        Chains.avalanche: 'AVALANCHE',
        Chains.tron: 'TRON',
        Chains.solana: 'SOLANA',
        Chains.base: 'BASE',
        Chains.klaytn: 'KLAYTN',
        Chains.ton: 'TON'
    }


NATIVE_TOKEN = '0x0000000000000000000000000000000000000000'

NATIVE_TOKENS = {
    Chains.bsc: '0x0000000000000000000000000000000000000000',
    Chains.ethereum: '0x0000000000000000000000000000000000000000',
    Chains.fantom: '0x0000000000000000000000000000000000000000',
    Chains.polygon: '0x0000000000000000000000000000000000000000',
    Chains.arbitrum: '0x0000000000000000000000000000000000000000',
    Chains.optimism: '0x0000000000000000000000000000000000000000',
    Chains.avalanche: '0x0000000000000000000000000000000000000000',
    Chains.tron: '0x0000000000000000000000000000000000000000',
    Chains.ton: 'EQAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAM9c'
}


class MulticallContract:
    """
    References: https://www.multicall3.com/deployments
    """

    on_chains_v3 = {
        Chains.ethereum: '0xca11bde05977b3631167028862be2a173976ca11',
        Chains.bsc: '0xca11bde05977b3631167028862be2a173976ca11',
        Chains.polygon: '0xca11bde05977b3631167028862be2a173976ca11',
        Chains.avalanche: '0xca11bde05977b3631167028862be2a173976ca11',
        Chains.fantom: '0xca11bde05977b3631167028862be2a173976ca11',
        Chains.arbitrum: '0xca11bde05977b3631167028862be2a173976ca11',
        Chains.optimism: '0xca11bde05977b3631167028862be2a173976ca11',
        Chains.tron: '0x32a4f47a74a6810bd0bf861cabab99656a75de9e',
        Chains.oasis_sapphire: '0xca11bde05977b3631167028862be2a173976ca11',
        Chains.oasis_sapphire_testnet: '0xca11bde05977b3631167028862be2a173976ca11'
    }

    default_address = '0xca11bde05977b3631167028862be2a173976ca11'

    @classmethod
    def get_multicall_contract(cls, chain_id):
        return cls.on_chains_v3.get(chain_id, cls.default_address)


EMPTY_TOKEN_IMG = 'https://firebasestorage.googleapis.com/v0/b/token-c515a.appspot.com/o/tokens_v2%2Fempty-token.png?alt=media&token=2f9dfcc1-88a0-472c-a51f-4babc0c583f0'



import json

COMET_ABI = json.loads('''
[
  {
    "inputs": [
      {
        "components": [
          {
            "internalType": "address",
            "name": "governor",
            "type": "address"
          },
          {
            "internalType": "address",
            "name": "pauseGuardian",
            "type": "address"
          },
          {
            "internalType": "address",
            "name": "baseToken",
            "type": "address"
          },
          {
            "internalType": "address",
            "name": "baseTokenPriceFeed",
            "type": "address"
          },
          {
            "internalType": "address",
            "name": "extensionDelegate",
            "type": "address"
          },
          {
            "internalType": "uint64",
            "name": "supplyKink",
            "type": "uint64"
          },
          {
            "internalType": "uint64",
            "name": "supplyPerYearInterestRateSlopeLow",
            "type": "uint64"
          },
          {
            "internalType": "uint64",
            "name": "supplyPerYearInterestRateSlopeHigh",
            "type": "uint64"
          },
          {
            "internalType": "uint64",
            "name": "supplyPerYearInterestRateBase",
            "type": "uint64"
          },
          {
            "internalType": "uint64",
            "name": "borrowKink",
            "type": "uint64"
          },
          {
            "internalType": "uint64",
            "name": "borrowPerYearInterestRateSlopeLow",
            "type": "uint64"
          },
          {
            "internalType": "uint64",
            "name": "borrowPerYearInterestRateSlopeHigh",
            "type": "uint64"
          },
          {
            "internalType": "uint64",
            "name": "borrowPerYearInterestRateBase",
            "type": "uint64"
          },
          {
            "internalType": "uint64",
            "name": "storeFrontPriceFactor",
            "type": "uint64"
          },
          {
            "internalType": "uint64",
            "name": "trackingIndexScale",
            "type": "uint64"
          },
          {
            "internalType": "uint64",
            "name": "baseTrackingSupplySpeed",
            "type": "uint64"
          },
          {
            "internalType": "uint64",
            "name": "baseTrackingBorrowSpeed",
            "type": "uint64"
          },
          {
            "internalType": "uint104",
            "name": "baseMinForRewards",
            "type": "uint104"
          },
          {
            "internalType": "uint104",
            "name": "baseBorrowMin",
            "type": "uint104"
          },
          {
            "internalType": "uint104",
            "name": "targetReserves",
            "type": "uint104"
          },
          {
            "components": [
              {
                "internalType": "address",
                "name": "asset",
                "type": "address"
              },
              {
                "internalType": "address",
                "name": "priceFeed",
                "type": "address"
              },
              {
                "internalType": "uint8",
                "name": "decimals",
                "type": "uint8"
              },
              {
                "internalType": "uint64",
                "name": "borrowCollateralFactor",
                "type": "uint64"
              },
              {
                "internalType": "uint64",
                "name": "liquidateCollateralFactor",
                "type": "uint64"
              },
              {
                "internalType": "uint64",
                "name": "liquidationFactor",
                "type": "uint64"
              },
              {
                "internalType": "uint128",
                "name": "supplyCap",
                "type": "uint128"
              }
            ],
            "internalType": "struct CometConfiguration.AssetConfig[]",
            "name": "assetConfigs",
            "type": "tuple[]"
          }
        ],
        "internalType": "struct CometConfiguration.Configuration",
        "name": "config",
        "type": "tuple"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "inputs": [],
    "name": "Absurd",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "AlreadyInitialized",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "BadAsset",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "BadDecimals",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "BadDiscount",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "BadMinimum",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "BadPrice",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "BorrowCFTooLarge",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "BorrowTooSmall",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "InsufficientReserves",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "InvalidInt104",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "InvalidInt256",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "InvalidUInt104",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "InvalidUInt128",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "InvalidUInt64",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "LiquidateCFTooLarge",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "NegativeNumber",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "NoSelfTransfer",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "NotCollateralized",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "NotForSale",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "NotLiquidatable",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "Paused",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "SupplyCapExceeded",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "TimestampTooLarge",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "TooManyAssets",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "TooMuchSlippage",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "TransferInFailed",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "TransferOutFailed",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "Unauthorized",
    "type": "error"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "absorber",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "borrower",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "collateralAbsorbed",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "usdValue",
        "type": "uint256"
      }
    ],
    "name": "AbsorbCollateral",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "absorber",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "borrower",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "basePaidOut",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "usdValue",
        "type": "uint256"
      }
    ],
    "name": "AbsorbDebt",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "buyer",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "baseAmount",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "collateralAmount",
        "type": "uint256"
      }
    ],
    "name": "BuyCollateral",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "bool",
        "name": "supplyPaused",
        "type": "bool"
      },
      {
        "indexed": false,
        "internalType": "bool",
        "name": "transferPaused",
        "type": "bool"
      },
      {
        "indexed": false,
        "internalType": "bool",
        "name": "withdrawPaused",
        "type": "bool"
      },
      {
        "indexed": false,
        "internalType": "bool",
        "name": "absorbPaused",
        "type": "bool"
      },
      {
        "indexed": false,
        "internalType": "bool",
        "name": "buyPaused",
        "type": "bool"
      }
    ],
    "name": "PauseAction",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "from",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "dst",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "Supply",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "from",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "dst",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "SupplyCollateral",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "from",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "to",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "Transfer",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "from",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "to",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "TransferCollateral",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "src",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "to",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "Withdraw",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "src",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "to",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "WithdrawCollateral",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "to",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "WithdrawReserves",
    "type": "event"
  },
  {
    "stateMutability": "payable",
    "type": "fallback"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "absorber",
        "type": "address"
      },
      {
        "internalType": "address[]",
        "name": "accounts",
        "type": "address[]"
      }
    ],
    "name": "absorb",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "accrueAccount",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "manager",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "approveThis",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "balanceOf",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "baseBorrowMin",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "baseMinForRewards",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "baseScale",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "baseToken",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "baseTokenPriceFeed",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "baseTrackingBorrowSpeed",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "baseTrackingSupplySpeed",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "borrowBalanceOf",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "borrowKink",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "borrowPerSecondInterestRateBase",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "borrowPerSecondInterestRateSlopeHigh",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "borrowPerSecondInterestRateSlopeLow",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "minAmount",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "baseAmount",
        "type": "uint256"
      },
      {
        "internalType": "address",
        "name": "recipient",
        "type": "address"
      }
    ],
    "name": "buyCollateral",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "decimals",
    "outputs": [
      {
        "internalType": "uint8",
        "name": "",
        "type": "uint8"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "extensionDelegate",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint8",
        "name": "i",
        "type": "uint8"
      }
    ],
    "name": "getAssetInfo",
    "outputs": [
      {
        "components": [
          {
            "internalType": "uint8",
            "name": "offset",
            "type": "uint8"
          },
          {
            "internalType": "address",
            "name": "asset",
            "type": "address"
          },
          {
            "internalType": "address",
            "name": "priceFeed",
            "type": "address"
          },
          {
            "internalType": "uint64",
            "name": "scale",
            "type": "uint64"
          },
          {
            "internalType": "uint64",
            "name": "borrowCollateralFactor",
            "type": "uint64"
          },
          {
            "internalType": "uint64",
            "name": "liquidateCollateralFactor",
            "type": "uint64"
          },
          {
            "internalType": "uint64",
            "name": "liquidationFactor",
            "type": "uint64"
          },
          {
            "internalType": "uint128",
            "name": "supplyCap",
            "type": "uint128"
          }
        ],
        "internalType": "struct CometCore.AssetInfo",
        "name": "",
        "type": "tuple"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "asset",
        "type": "address"
      }
    ],
    "name": "getAssetInfoByAddress",
    "outputs": [
      {
        "components": [
          {
            "internalType": "uint8",
            "name": "offset",
            "type": "uint8"
          },
          {
            "internalType": "address",
            "name": "asset",
            "type": "address"
          },
          {
            "internalType": "address",
            "name": "priceFeed",
            "type": "address"
          },
          {
            "internalType": "uint64",
            "name": "scale",
            "type": "uint64"
          },
          {
            "internalType": "uint64",
            "name": "borrowCollateralFactor",
            "type": "uint64"
          },
          {
            "internalType": "uint64",
            "name": "liquidateCollateralFactor",
            "type": "uint64"
          },
          {
            "internalType": "uint64",
            "name": "liquidationFactor",
            "type": "uint64"
          },
          {
            "internalType": "uint128",
            "name": "supplyCap",
            "type": "uint128"
          }
        ],
        "internalType": "struct CometCore.AssetInfo",
        "name": "",
        "type": "tuple"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "utilization",
        "type": "uint256"
      }
    ],
    "name": "getBorrowRate",
    "outputs": [
      {
        "internalType": "uint64",
        "name": "",
        "type": "uint64"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "asset",
        "type": "address"
      }
    ],
    "name": "getCollateralReserves",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "priceFeed",
        "type": "address"
      }
    ],
    "name": "getPrice",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getReserves",
    "outputs": [
      {
        "internalType": "int256",
        "name": "",
        "type": "int256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "utilization",
        "type": "uint256"
      }
    ],
    "name": "getSupplyRate",
    "outputs": [
      {
        "internalType": "uint64",
        "name": "",
        "type": "uint64"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getUtilization",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "governor",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "owner",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "manager",
        "type": "address"
      }
    ],
    "name": "hasPermission",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "initializeStorage",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "isAbsorbPaused",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "isAllowed",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "isBorrowCollateralized",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "isBuyPaused",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "isLiquidatable",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "isSupplyPaused",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "isTransferPaused",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "isWithdrawPaused",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "liquidatorPoints",
    "outputs": [
      {
        "internalType": "uint32",
        "name": "numAbsorbs",
        "type": "uint32"
      },
      {
        "internalType": "uint64",
        "name": "numAbsorbed",
        "type": "uint64"
      },
      {
        "internalType": "uint128",
        "name": "approxSpend",
        "type": "uint128"
      },
      {
        "internalType": "uint32",
        "name": "_reserved",
        "type": "uint32"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "numAssets",
    "outputs": [
      {
        "internalType": "uint8",
        "name": "",
        "type": "uint8"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bool",
        "name": "supplyPaused",
        "type": "bool"
      },
      {
        "internalType": "bool",
        "name": "transferPaused",
        "type": "bool"
      },
      {
        "internalType": "bool",
        "name": "withdrawPaused",
        "type": "bool"
      },
      {
        "internalType": "bool",
        "name": "absorbPaused",
        "type": "bool"
      },
      {
        "internalType": "bool",
        "name": "buyPaused",
        "type": "bool"
      }
    ],
    "name": "pause",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "pauseGuardian",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "baseAmount",
        "type": "uint256"
      }
    ],
    "name": "quoteCollateral",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "storeFrontPriceFactor",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "supply",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "from",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "dst",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "supplyFrom",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "supplyKink",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "supplyPerSecondInterestRateBase",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "supplyPerSecondInterestRateSlopeHigh",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "supplyPerSecondInterestRateSlopeLow",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "dst",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "supplyTo",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "targetReserves",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "totalBorrow",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "totalSupply",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "totalsCollateral",
    "outputs": [
      {
        "internalType": "uint128",
        "name": "totalSupplyAsset",
        "type": "uint128"
      },
      {
        "internalType": "uint128",
        "name": "_reserved",
        "type": "uint128"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "trackingIndexScale",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "dst",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "transfer",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "dst",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "transferAsset",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "src",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "dst",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "transferAssetFrom",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "src",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "dst",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "transferFrom",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "userBasic",
    "outputs": [
      {
        "internalType": "int104",
        "name": "principal",
        "type": "int104"
      },
      {
        "internalType": "uint64",
        "name": "baseTrackingIndex",
        "type": "uint64"
      },
      {
        "internalType": "uint64",
        "name": "baseTrackingAccrued",
        "type": "uint64"
      },
      {
        "internalType": "uint16",
        "name": "assetsIn",
        "type": "uint16"
      },
      {
        "internalType": "uint8",
        "name": "_reserved",
        "type": "uint8"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "userCollateral",
    "outputs": [
      {
        "internalType": "uint128",
        "name": "balance",
        "type": "uint128"
      },
      {
        "internalType": "uint128",
        "name": "_reserved",
        "type": "uint128"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "name": "userNonce",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "withdraw",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "src",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "to",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "withdrawFrom",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "to",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "withdrawReserves",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "to",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "asset",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "withdrawTo",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  }
]
''')


import copy
import inspect


def flatten_dict(d):
    out = {}
    for key, val in d.items():
        if isinstance(val, dict):
            val = [val]
        if isinstance(val, list):
            array = []
            for subdict in val:
                if not isinstance(subdict, dict):
                    array.append(subdict)
                else:
                    deeper = flatten_dict(subdict).items()
                    out.update({str(key) + '.' + str(key2): val2 for key2, val2 in deeper})
            if array:
                out.update({str(key): array})
        else:
            out[str(key)] = val
    return out


def reverse_flatten_dict(d: dict) -> dict:
    result = {}
    for key, val in d.items():
        nested_keys = key.split('.')
        d_ = result
        for k in nested_keys[:-1]:
            if k not in d_:
                d_[k] = {}
            d_ = d_[k]
        d_[nested_keys[-1]] = val

    return result


def add_dict(first_dict: dict, second_dict: dict):
    ans_dict = dict()
    for key in first_dict.keys():
        ans_dict[key] = first_dict[key] + second_dict[key]
    return ans_dict


def remove_none_value_dict(a_dict):
    copy_dict = a_dict.copy()
    for key, value in copy_dict.items():
        if value is None:
            a_dict.pop(key)
    return a_dict


def delete_none(_dict):
    """Delete None values recursively from all the dictionaries"""
    for key, value in list(_dict.items()):
        if isinstance(value, dict):
            delete_none(value)
        elif value is None:
            del _dict[key]
        elif isinstance(value, list):
            for v_i in value:
                if isinstance(v_i, dict):
                    delete_none(v_i)

    return _dict


def filter_doc_by_keys(doc, keys):
    doc_ = copy.deepcopy(doc)
    if keys is None:
        return doc_
    return dict(filter(lambda x: x[0] in keys, doc_.items()))


def get_class_constant_keys(cls):
    keys = []
    for attr in inspect.getmembers(cls):
        # to remove private and protected functions
        if not attr[0].startswith('_'):
            # To remove function
            if not inspect.ismethod(attr[1]):
                keys.append(attr[1])

    return keys


def short_address(address: str, n_start=3, n_end=4):
    if not address:
        return ''
    return address[:n_start + 2] + '...' + address[-n_end:]


def all_equal(array):
    return (not array) or (array.count(array[0]) == len(array))




import json

ACCESS_CONTROLLED_V8 = json.loads("""
[
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "sender",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "calledContract",
        "type": "address"
      },
      {
        "internalType": "string",
        "name": "methodSignature",
        "type": "string"
      }
    ],
    "name": "Unauthorized",
    "type": "error"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint8",
        "name": "version",
        "type": "uint8"
      }
    ],
    "name": "Initialized",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "address",
        "name": "oldAccessControlManager",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "address",
        "name": "newAccessControlManager",
        "type": "address"
      }
    ],
    "name": "NewAccessControlManager",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "previousOwner",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "newOwner",
        "type": "address"
      }
    ],
    "name": "OwnershipTransferStarted",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "previousOwner",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "newOwner",
        "type": "address"
      }
    ],
    "name": "OwnershipTransferred",
    "type": "event"
  },
  {
    "inputs": [],
    "name": "acceptOwnership",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "accessControlManager",
    "outputs": [
      {
        "internalType": "contract IAccessControlManagerV8",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "owner",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "pendingOwner",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "renounceOwnership",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "accessControlManager_",
        "type": "address"
      }
    ],
    "name": "setAccessControlManager",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "newOwner",
        "type": "address"
      }
    ],
    "name": "transferOwnership",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  }
]
""")
        